from swmmio.defs.config import FONT_PATH

class _dotdict(dict):
    """dot.notation access to dictionary attributes"""
    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

"""
this allows module wide configuration of drawing methods. The dotdict allows for
convenient access.
Example:

    from swmmio import graphics
    from swmmio.graphics import swmm_graphics as sg

    #configure
    graphics.options.inlcude_parcels = True

    #draws the model with parcels
    sg.drawModel(swmmio.Model(/path/to/model), bbox=su.d68d70)

"""
_o = {
    'include_basemap':False,
    'include_parcels':False,
    'basemap_shapefile_dir':r'C:\Data\ArcGIS\Shapefiles',

    #regular shapefile used for drawing parcels
    'parcels_shapefile':r'C:\Data\ArcGIS\Shapefiles\pennsport_parcels.shp',

    #table resulting from one-to-many spatial join of parcels to sheds
    'parcel_node_join_data':r'P:\02_Projects\SouthPhila\SE_SFR\MasterModels\CommonData\pennsport_sheds_parcels_join.csv',

    'font_file':FONT_PATH,#r'C:\Data\Code\Fonts\Raleway-Regular.ttf',
}

config = _dotdict(_o)



from swmmio.tests.data import (MODEL_XSECTION_BASELINE, MODEL_XSECTION_ALT_01,
                               MODEL_XSECTION_ALT_02, MODEL_XSECTION_ALT_03)
from swmmio import swmmio
from swmmio.version_control import utils as vc_utils
from swmmio.version_control import inp
from swmmio.utils import functions as funcs


def test_complete_inp_headers():

    headers = [
        '[TITLE]', '[OPTIONS]', '[EVAPORATION]', '[JUNCTIONS]', '[OUTFALLS]',
        '[CONDUITS]', '[XSECTIONS]', '[DWF]', '[REPORT]', '[TAGS]', '[MAP]',
        '[COORDINATES]', '[VERTICES]',
    ]

    h1 = funcs.complete_inp_headers(MODEL_XSECTION_BASELINE)

    assert(all(h in h1['headers'] for h in headers))
    assert(h1['order'] == headers)


def test_create_inp_build_instructions():

    inp.create_inp_build_instructions(MODEL_XSECTION_BASELINE,
                                      MODEL_XSECTION_ALT_03,
                                      'vc_dir',
                                      'test_version_id', 'cool comments')

    latest_bi = vc_utils.newest_file('vc_dir')
    bi = inp.BuildInstructions(latest_bi)

    juncs = bi.instructions['[JUNCTIONS]']
    assert(all(j in juncs.altered.index for j in [
           'dummy_node1', 'dummy_node5']))



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2018 Adam Erispaha
#
# Licensed under the terms of the BSD2 License
# See LICENSE.txt for details
# -----------------------------------------------------------------------------
"""SWMM5 test models."""

# Standard library imports
import os

DATA_PATH = os.path.abspath(os.path.dirname(__file__))

# Test models paths
MODEL_FULL_FEATURES_PATH = os.path.join(DATA_PATH, 'model_full_features.inp')
MODEL_FULL_FEATURES_XY = os.path.join(
    DATA_PATH, 'model_full_features_network_xy.inp')
MODEL_FULL_FEATURES__NET_PATH = os.path.join(
    DATA_PATH, 'model_full_features_network.inp')
MODEL_BROWARD_COUNTY_PATH = os.path.join(DATA_PATH, 'RUNOFF46_SW5.INP')

# version control test models
MODEL_XSECTION_BASELINE = os.path.join(DATA_PATH, 'baseline_test.inp')
MODEL_XSECTION_ALT_01 = os.path.join(DATA_PATH, 'alt_test1.inp')
MODEL_XSECTION_ALT_02 = os.path.join(DATA_PATH, 'alt_test2.inp')
MODEL_XSECTION_ALT_03 = os.path.join(DATA_PATH, 'alt_test3.inp')

df_test_coordinates_csv = os.path.join(DATA_PATH, 'df_test_coordinates.csv')



from swmmio.defs.sectionheaders import inp_header_dict, rpt_header_dict
from collections import deque
import pandas as pd


def random_alphanumeric(n=6):
    import random
    chars = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz'
    return ''.join(random.choice(chars) for i in range(n))


def model_to_networkx(model, drop_cycles=True):
    from swmmio.utils.dataframes import create_dataframeINP, create_dataframeRPT
    '''
    Networkx MultiDiGraph representation of the model
    '''
    from geojson import Point, LineString
    try:
        import networkx as nx
    except ImportError:
        raise ImportError('networkx module needed. get this package here: ',
                        'https://pypi.python.org/pypi/networkx')

    def multidigraph_from_edges(edges, source, target):
        '''
        create a MultiDiGraph from a dataframe of edges, using the row index
        as the key in the MultiDiGraph
        '''
        us = edges[source]
        vs = edges[target]
        keys = edges.index
        data = edges.drop([source, target], axis=1)
        d_dicts = data.to_dict(orient='records')

        G = nx.MultiDiGraph()

        G.add_edges_from(zip(us, vs, keys, d_dicts))

        return G

    # parse swmm model results with swmmio, concat all links into one dataframe
    nodes = model.nodes()
    if model.rpt is not None:
        inflow_cols = [
            'MaxLatInflow',
            'MaxTotalInflow',
            'LatInflowV',
            'TotalInflowV',
            'FlowBalErrorPerc']
        flows = create_dataframeRPT(
            model.rpt.path, "Node Inflow Summary")[inflow_cols]
        nodes = nodes.join(flows)

    conduits = model.conduits()
    links = pd.concat([conduits, model.orifices(), model.weirs(), model.pumps()])
    links['facilityid'] = links.index

    # create a nx.MultiDiGraph from the combined model links, add node data, set CRS
    G = multidigraph_from_edges(links, 'InletNode', target='OutletNode')
    G.add_nodes_from(zip(nodes.index, nodes.to_dict(orient='records')))

    # create geojson geometry objects for each graph element
    for u, v, k, coords in G.edges(data='coords', keys=True):
        if coords:
            G[u][v][k]['geometry'] = LineString(coords)
    for n, coords in G.nodes(data='coords'):
        if coords:
            G.node[n]['geometry'] = Point(coords[0])

    if drop_cycles:
        # remove cycles
        cycles = list(nx.simple_cycles(G))
        if len(cycles) > 0:
            print('cycles detected and removed: {}'.format(cycles))
            G.remove_edges_from(cycles)

    G.graph['crs'] = model.crs
    return G


# Todo: use an OrderedDict instead of a dict and a "order" list
def complete_inp_headers(inpfilepath):
    """
    creates a dictionary with all the headers found in an INP file
    (which varies based on what the user has defined in a given model)
    and updates them based on the definitions in inp_header_dict
    this ensures the list is comprehensive

    RETURNS:
        a dictionary including
            'headers'->
                    header section keys and their respective cleaned column headers
            'order' ->
                    an array of section headers found in the INP file
                    that preserves the original order
    """
    foundheaders = {}
    order = []
    # print inp_header_dict
    with open(inpfilepath) as f:
        for line in f:
            if '[' and ']' in line:
                h = line.strip()
                order.append(h)
                if h in inp_header_dict:
                    foundheaders.update({h: inp_header_dict[h]})
                else:
                    foundheaders.update({h: 'blob'})

    return {'headers': foundheaders, 'order': order}


def complete_rpt_headers(rptfilepath):
    """
    creates a dictionary with all the headers found in an RPT file
    (which varies based on what the user has defined in a given model)
    and updates them based on the definitions in rpt_header_dict
    this ensures the list is comprehensive

    RETURNS:
        a dictionary including
            'headers'->
                    header section keys and their respective cleaned column headers
            'order' ->
                    an array of section headers found in the RPT file
                    that perserves the original order
    """
    foundheaders = {}
    order = []
    with open(rptfilepath) as f:
        buff3line = deque()
        for line in f:


            #maintains a 3 line buffer and looks for instances where
            #a top and bottom line have '*****' and records the middle line
            #typical of section headers in RPT files
            buff3line.append(line)
            if len(buff3line) > 3:
                buff3line.popleft()

            if ('***********'in buff3line[0] and
                '***********'in buff3line[2] and
                len(buff3line[1].strip()) > 0):
                h = buff3line[1].strip()
                order.append(h)
                if h in rpt_header_dict:
                    foundheaders.update({h:rpt_header_dict[h]})
                else:
                    foundheaders.update({h:'blob'})

    return {'headers':foundheaders, 'order':order}


def merge_dicts(*dict_args):
    '''
    Given any number of dicts, shallow copy and merge into a new dict,
    precedence goes to key value pairs in latter dicts.
    '''
    result = {}
    for dictionary in dict_args:
        if dictionary:
            result.update(dictionary)
    return result


def trace_from_node(conduits, startnode, mode='up', stopnode=None):

    """
    trace up and down a SWMM model given a start node and optionally a
    stop node.
    """

    traced_nodes = [startnode] #include the starting node
    traced_conduits = []

    def trace(node_id):
        for conduit, data in conduits.iterrows():
            if mode == 'up' and data.OutletNode == node_id and conduit not in traced_conduits:

                traced_nodes.append(data.InletNode)
                traced_conduits.append(conduit)

                if stopnode and data.InletNode == stopnode:
                    break
                trace(data.InletNode)

            if mode == 'down' and data.InletNode == node_id and conduit not in traced_conduits:
                traced_nodes.append(data.OutletNode)
                traced_conduits.append(conduit)

                if stopnode and data.OutletNode == stopnode:
                    break
                trace(data.OutletNode)

    #kickoff the trace
    print ("Starting trace {} from {}".format(mode, startnode))
    trace(startnode)
    print ("Traced {0} nodes from {1}".format(len(traced_nodes), startnode))
    return {'nodes':traced_nodes, 'conduits':traced_conduits}



import pandas as pd
import shutil
import os
import fileinput
import itertools
from datetime import datetime
from swmmio.swmmio import Model
from swmmio.utils import functions as funcs
from swmmio.utils.dataframes import create_dataframeINP
# from swmmio.utils import swmm_utils as su
from swmmio.version_control import utils as vc_utils
from swmmio.version_control import inp

#from .utils.text import * #functions for processing inp/rpt/txt files

pd.options.display.max_colwidth = 200


def propagate_changes_from_baseline(baseline_dir, alternatives_dir, combi_dir,
                                    version_id='', comments=''):

    #stuff
    """
    if the baseline model has changes that need to be propogated to all models,
    iterate through each model and rebuild the INPs with the new baseline and
    existing build instructions. update the build instructions to reflect the
    revision date of the baseline.
    """
    version_id += '_' + datetime.now().strftime("%y%m%d%H%M%S")

    #collect the directories of all models
    model_dirs = []
    for alt in os.listdir(alternatives_dir):
        #print alt
        #iterate through each implementation level of each alternative
        for imp_level in os.listdir(os.path.join(alternatives_dir, alt)):
            #create or refresh the build instructions file for the alternatives
            model_dirs.append(os.path.join(alternatives_dir, alt, imp_level))

    model_dirs += [os.path.join(combi_dir, x) for x in os.listdir(combi_dir)]
    #print model_dirs
    baseline = Model(baseline_dir)
    baseinp = baseline.inp.path

    for model_dir in model_dirs:
        model = Model(model_dir)
        vc_directory = os.path.join(model_dir, 'vc')
        latest_bi = vc_utils.newest_file(vc_directory)

        #update build instructions metdata and build the new inp
        bi = inp.BuildInstructions(latest_bi)
        bi.metadata['Parent Models']['Baseline'] = {baseinp:vc_utils.modification_date(baseinp)}
        bi.metadata['Log'].update({version_id:comments})
        bi.save(vc_directory, version_id+'.txt')
        print('rebuilding {} with changes to baseline'.format(model.name))
        bi.build(baseline_dir, model.inp.path) #overwrite old inp




def create_combinations(baseline_dir, rsn_dir, combi_dir, version_id='',
                        comments=''):

    """
    Generate SWMM5 models of each logical combination of all implementation
    phases (IP) across all relief sewer networks (RSN).

    Inputs:
        baseline_dir -> path to directory containing the baseline SWMM5 model
        rsn_dir ->      path to directory containing subdirectories for each RSN
                        containing directories for each IP within the network
        combi_dir ->    target directory in which child models will be created
        version_id ->   identifier for a given version (optional)
        comments ->     comments tracked within build instructions log for
                        each model scenario (optional)

    Calling create_combinations will update child models if parent models have
    been changed.

    """

    baseinp = Model(baseline_dir).inp.path
    version_id += '_' + datetime.now().strftime("%y%m%d%H%M%S")

    #create a list of directories pointing to each IP in each RSN
    RSN_dirs = [os.path.join(rsn_dir, rsn) for rsn in os.listdir(rsn_dir)]
    IP_dirs = [os.path.join(d, ip) for d in RSN_dirs for ip in os.listdir(d)]

    #list of lists of each IP within each RSN, including a 'None' phase.
    IPs = [[None] + os.listdir(d) for d in RSN_dirs]

    #identify all scenarios (cartesian product of sets of IPs between each RSN)
    #then isolate child scenarios with atleast 2 parents (sets with one parent
    #are already modeled as IPs within the RSNs)
    all_scenarios = [[_f for _f in s if _f] for s in itertools.product(*IPs)]
    child_scenarios = [s for s in all_scenarios if len(s) > 1]

    #notify user of what was initially found
    str_IPs = '\n'.join([', '.join([_f for _f in i if _f]) for i in IPs])
    print(('Found {} implementation phases among {} networks:\n{}\n'
           'This yeilds {} combined scenarios ({} total)'.format(len(IP_dirs),
            len(RSN_dirs),str_IPs,len(child_scenarios),len(all_scenarios) - 1)))

    # ==========================================================================
    # UPDATE/CREATE THE PARENT MODEL BUILD INSTRUCTIONS
    # ==========================================================================
    for ip_dir in IP_dirs:
        ip_model = Model(ip_dir)
        vc_dir = os.path.join(ip_dir, 'vc')

        if not os.path.exists(vc_dir):
            print('creating new build instructions for {}'.format(ip_model.name))
            inp.create_inp_build_instructions(baseinp, ip_model.inp.path,
                                              vc_dir,
                                              version_id, comments)
        else:
            #check if the alternative model was changed since last run of this tool
            #--> compare the modification date to the BI's modification date meta data
            latest_bi = vc_utils.newest_file(vc_dir)
            if not vc_utils.bi_is_current(latest_bi):
                #revision date of the alt doesn't match the newest build
                #instructions for this 'imp_level', so we should refresh it
                print('updating build instructions for {}'.format(ip_model.name))
                inp.create_inp_build_instructions(baseinp, ip_model.inp.path,
                                                  vc_dir, version_id,
                                                  comments)

    # ==========================================================================
    # UPDATE/CREATE THE CHILD MODELS AND CHILD BUILD INSTRUCTIONS
    # ==========================================================================
    for scen in child_scenarios:
        newcombi = '_'.join(sorted(scen))
        new_dir = os.path.join(combi_dir, newcombi)
        vc_dir = os.path.join(combi_dir, newcombi, 'vc')

        #parent model build instr files
        #BUG (this breaks with model IDs with more than 1 char)
        parent_vc_dirs = [os.path.join(rsn_dir, f[0], f, 'vc') for f in scen]
        latest_parent_bis = [vc_utils.newest_file(d) for d in parent_vc_dirs]
        build_instrcts = [inp.BuildInstructions(bi) for bi in latest_parent_bis]

        if not os.path.exists(new_dir):

            os.mkdir(new_dir)
            newinppath = os.path.join(new_dir, newcombi + '.inp')

            print('creating new child model: {}'.format(newcombi))
            new_build_instructions = sum(build_instrcts)
            new_build_instructions.save(vc_dir, version_id+'.txt')
            new_build_instructions.build(baseline_dir, newinppath)

        else:
            #check if the alternative model was changed since last run
            #of this tool --> compare the modification date to the BI's
            #modification date meta data
            latest_bi = vc_utils.newest_file(os.path.join(new_dir,'vc'))
            if not vc_utils.bi_is_current(latest_bi):
                #revision date of the alt doesn't match the newest build
                #instructions for this 'imp_level', so we should refresh it
                print('updating child build instructions for {}'.format(newcombi))
                newinppath = os.path.join(new_dir, newcombi + '.inp')
                new_build_instructions = sum(build_instrcts)
                new_build_instructions.save(vc_dir, version_id+'.txt')
                new_build_instructions.build(baseline_dir, newinppath)



#!/usr/bin/env python
# coding:utf-8
import re
from time import ctime
import pandas as pd
import glob
import math
from swmmio.utils import spatial
from swmmio.utils import functions
from swmmio.utils.dataframes import create_dataframeINP, create_dataframeRPT, get_link_coords
from swmmio.defs.config import *
from swmmio.tests.data import MODEL_FULL_FEATURES__NET_PATH, MODEL_FULL_FEATURES_XY
import warnings
import swmmio
from swmmio.elements import ModelSection


class Model(object):

    def __init__(self, in_file_path, crs=None):
        """
        Class representing a complete SWMM model incorporating its INP and RPT
        files and data

        initialize a swmmio.Model object by pointing it to a directory containing
        a single INP (and optionally an RPT file with matching filename) or by
        pointing it directly to an .inp file.
        """
        self.crs = None
        inp_path = None
        if os.path.isdir(in_file_path):
            # a directory was passed in
            inps_in_dir = glob.glob1(in_file_path, "*.inp")
            if len(inps_in_dir) == 1:
                # there is only one INP in this directory -> good.
                inp_path = os.path.join(in_file_path, inps_in_dir[0])

        elif os.path.splitext(in_file_path)[1] == '.inp':
            # an inp was passed in
            inp_path = in_file_path

        if inp_path:
            wd = os.path.dirname(inp_path)  # working dir
            name = os.path.splitext(os.path.basename(inp_path))[0]
            self.name = name
            self.inp = inp(inp_path)  # inp object
            self.rpt = None  # until we can confirm it initializes properly
            self.bbox = None  # to remember how the model data was clipped
            self.scenario = ''  # self._get_scenario()
            self.crs = crs  # coordinate reference system

            # try to initialize a companion RPT object
            rpt_path = os.path.join(wd, name + '.rpt')
            if os.path.exists(rpt_path):
                try:
                    self.rpt = rpt(rpt_path)
                except Exception as e:
                    print('{}.rpt failed to initialize\n{}'.format(name, e))

            self._nodes_df = None
            self._conduits_df = None
            self._orifices_df = None
            self._weirs_df = None
            self._pumps_df = None
            self._links_df = None
            self._subcatchments_df = None
            self._network = None

    def rpt_is_valid(self, verbose=False):
        """
        Return true if the .rpt file exists and has a revision date more
        recent than the .inp file. If the inp has an modified date later than
        the rpt, assume that the rpt should be regenerated
        """

        if self.rpt is None:
            if verbose:
                print('{} does not have an rpt file'.format(self.name))
            return False

        # check if the rpt has ERRORS output from SWMM
        with open(self.rpt.path) as f:
            # jump to 500 bytes before the end of file
            f.seek(self.rpt.file_size - 500)
            for line in f:
                spl = line.split()
                if len(spl) > 0 and spl[0] == 'ERROR':
                    # return false at first "ERROR" occurence
                    return False

        rpt_mod_time = os.path.getmtime(self.rpt.path)
        inp_mod_time = os.path.getmtime(self.inp.path)

        if verbose:
            print("{}.rpt: modified {}".format(self.name, ctime(rpt_mod_time)))
            print("{}.inp: modified {}".format(self.name, ctime(inp_mod_time)))

        if inp_mod_time > rpt_mod_time:
            # inp datetime modified greater than rpt datetime modified
            return False
        else:
            return True

    def to_map(self, filename=None, inproj='epsg:2272'):
        '''
        To be removed in v0.4.0. Use swmmio.reporting.visualize.create_map()
        '''

        def wrn():
            w = '''to_map is no longer supported! Use
            swmmio.reporting.visualize.create_map() instead'''
            warnings.warn(w, DeprecationWarning)

        with warnings.catch_warnings():
            warnings.simplefilter("always")
            wrn()

    def conduits(self):
        """
        collect all useful and available data related model conduits and
        organize in one dataframe.
        """

        # check if this has been done already and return that data accordingly
        if self._conduits_df is not None:
            return self._conduits_df

        # parse out the main objects of this model
        inp = self.inp
        rpt = self.rpt

        # create dataframes of relevant sections from the INP
        conduits_df = create_dataframeINP(inp.path, "[CONDUITS]", comment_cols=False)
        xsections_df = create_dataframeINP(inp.path, "[XSECTIONS]", comment_cols=False)
        conduits_df = conduits_df.join(xsections_df)

        if rpt:
            # create a dictionary holding data from an rpt file, if provided
            link_flow_df = create_dataframeRPT(rpt.path, "Link Flow Summary")
            conduits_df = conduits_df.join(link_flow_df)

        # add conduit coordinates
        xys = conduits_df.apply(lambda r: get_link_coords(r, self.inp.coordinates, self.inp.vertices), axis=1)
        df = conduits_df.assign(coords=xys.map(lambda x: x[0]))

        # add conduit up/down inverts and calculate slope
        elevs = self.nodes()[['InvertElev']]
        df = pd.merge(df, elevs, left_on='InletNode', right_index=True, how='left')
        df = df.rename(index=str, columns={"InvertElev": "InletNodeInvert"})
        df = pd.merge(df, elevs, left_on='OutletNode', right_index=True, how='left')
        df = df.rename(index=str, columns={"InvertElev": "OutletNodeInvert"})
        df['UpstreamInvert'] = df.InletNodeInvert + df.InletOffset
        df['DownstreamInvert'] = df.OutletNodeInvert + df.OutletOffset
        df['SlopeFtPerFt'] = (df.UpstreamInvert - df.DownstreamInvert) / df.Length

        df.InletNode = df.InletNode.astype(str)
        df.OutletNode = df.OutletNode.astype(str)

        self._conduits_df = df

        return df

    def orifices(self):
        """
        collect all useful and available data related model orifices and
        organize in one dataframe.
        """

        # check if this has been done already and return that data accordingly
        if self._orifices_df is not None:
            return self._orifices_df

        # parse out the main objects of this model
        inp = self.inp
        rpt = self.rpt

        # create dataframes of relevant sections from the INP
        orifices_df = create_dataframeINP(inp.path, "[ORIFICES]", comment_cols=False)
        if orifices_df.empty:
            return pd.DataFrame()

        # add conduit coordinates
        xys = orifices_df.apply(lambda r: get_link_coords(r, self.inp.coordinates, self.inp.vertices), axis=1)
        df = orifices_df.assign(coords=xys.map(lambda x: x[0]))
        df.InletNode = df.InletNode.astype(str)
        df.OutletNode = df.OutletNode.astype(str)
        self._orifices_df = df

        return df

    def weirs(self):
        """
        collect all useful and available data related model weirs and
        organize in one dataframe.
        """

        # check if this has been done already and return that data accordingly
        if self._weirs_df is not None:
            return self._weirs_df

        # parse out the main objects of this model
        inp = self.inp
        rpt = self.rpt

        # create dataframes of relevant sections from the INP
        weirs_df = create_dataframeINP(inp.path, "[WEIRS]")
        if weirs_df.empty:
            return pd.DataFrame()

        weirs_df = weirs_df[['InletNode', 'OutletNode', 'WeirType', 'CrestHeight']]

        # add conduit coordinates
        xys = weirs_df.apply(lambda r: get_link_coords(r, self.inp.coordinates, self.inp.vertices), axis=1)
        df = weirs_df.assign(coords=xys.map(lambda x: x[0]))
        df.InletNode = df.InletNode.astype(str)
        df.OutletNode = df.OutletNode.astype(str)

        self._weirs_df = df

        return df

    @property
    def pumps(self):

        """
        collect all useful and available data related model pumps and
        organize in one dataframe.
        >>> import swmmio
        >>> from swmmio.tests.data import MODEL_FULL_FEATURES_XY
        >>> model = swmmio.Model(MODEL_FULL_FEATURES_XY)
        >>> pumps = model.pumps()
        >>> pumps[['PumpCurve', 'InitStatus']]
             PumpCurve InitStatus
        Name
        C2    P1_Curve         ON
        >>> pumps = model.pumps.to_gdf()
        >>> pumps
        """

        pumps_df = ModelSection(self, 'pumps')
        self._pumps_df = pumps_df

        return pumps_df

    def links(self):
        """
        create a DataFrame containing all link objects in the model including conduits, pumps, weirs, and orifices.
        :return: dataframe containing all link objects in the model
        """
        if self._links_df is not None:
            return self._links_df

        df = pd.concat([self.conduits(), self.orifices(), self.weirs(), self.pumps()])
        df['facilityid'] = df.index
        self._links_df = df
        return df

    def nodes(self, bbox=None, subset=None):
        """
        collect all useful and available data related model nodes and organize
        in one dataframe.
        """

        # check if this has been done already and return that data accordingly
        if self._nodes_df is not None and bbox == self.bbox:
            return self._nodes_df

        # parse out the main objects of this model
        inp = self.inp
        rpt = self.rpt

        # create dataframes of relevant sections from the INP
        juncs_df = create_dataframeINP(inp.path, "[JUNCTIONS]")
        outfalls_df = create_dataframeINP(inp.path, "[OUTFALLS]")
        storage_df = create_dataframeINP(inp.path, "[STORAGE]")

        # concatenate the DFs and keep only relevant cols
        all_nodes = pd.concat([juncs_df, outfalls_df, storage_df])
        cols = ['InvertElev', 'MaxDepth', 'SurchargeDepth', 'PondedArea']
        all_nodes = all_nodes[cols]

        if rpt:
            # add results data if a rpt file was found
            depth_summ = create_dataframeRPT(rpt.path, "Node Depth Summary")
            flood_summ = create_dataframeRPT(rpt.path, "Node Flooding Summary")

            # join the rpt data (index on depth df, suffixes for common cols)
            rpt_df = depth_summ.join(
                flood_summ, lsuffix='_depth', rsuffix='_flood')
            all_nodes = all_nodes.join(rpt_df)  # join to the all_nodes df

        all_nodes = all_nodes.join(self.inp.coordinates[['X', 'Y']])

        def nodexy(row):
            if math.isnan(row.X) or math.isnan(row.Y):
                return None
            else:
                return [(row.X, row.Y)]

        xys = all_nodes.apply(lambda r: nodexy(r), axis=1)
        all_nodes = all_nodes.assign(coords=xys)
        all_nodes = all_nodes.rename(index=str)
        self._nodes_df = all_nodes

        return all_nodes

    def subcatchments(self):
        """
        collect all useful and available data related subcatchments and organize
        in one dataframe.
        """
        subs = create_dataframeINP(self.inp.path, "[SUBCATCHMENTS]")
        subs = subs.drop([';', 'Comment', 'Origin'], axis=1)
        polygons_df = self.inp.polygons

        if self.rpt:
            flw = create_dataframeRPT(
                self.rpt.path, 'Subcatchment Runoff Summary')
            subs = subs.join(flw)

            # more accurate runoff calculations
            subs['RunoffAcFt'] = subs.TotalRunoffIn / 12.0 * subs.Area
            subs['RunoffMGAccurate'] = subs.RunoffAcFt / 3.06888785

        self._subcatchments_df = subs

        return subs

    def node(self, node, conduit=None):
        '''
        To be removed in v0.4.0
        '''

        def wrn():
            w = "Depreciated. Use model.nodes().loc['{}'] instead".format(node)
            warnings.warn(w, DeprecationWarning)
            return self.nodes().loc[node]

        with warnings.catch_warnings():
            warnings.simplefilter("always")
            wrn()

    @property
    def network(self):
        """
        Networkx MultiDiGraph representation of the model
        :return: Networkx MultiDiGraph representation of model
        :rtype: networkx.MultiDiGraph
        """

        if self._network is None:
            G = functions.model_to_networkx(self, drop_cycles=False)
            self._network = G

        return self._network

    def to_crs(self, *args, **kwargs):
        """
        Convert coordinate reference system of the model coordinates
        :param target_crs:
        :return: True
        Example:
        >>> import swmmio
        >>> m = swmmio.Model(MODEL_FULL_FEATURES_XY, crs="+init=EPSG:2272")
        >>> m.to_crs("+init=EPSG:4326") # convert to WGS84 web mercator
        >>> m.inp.coordinates
                      X          Y
        Name                      
        J3   -74.866424  42.365958
        1    -74.870614  42.368292
        2    -74.867615  42.367916
        3    -74.869387  42.368527
        4    -74.869024  42.368089
        5    -74.868888  42.367709
        J2   -74.868458  42.366748
        J4   -74.864787  42.365966
        J1   -74.868861  42.366968
        >>> m.inp.vertices
                       X          Y
        Name                       
        C1:C2 -74.868703  42.366833
        C2.1  -74.868034  42.366271
        C2.1  -74.867305  42.365974
        """
        try:
            import pyproj
        except ImportError:
            raise ImportError('pyproj module needed. get this package here: ',
                              'https://pypi.python.org/pypi/pyproj')

        if self.crs is None:
            raise AttributeError('CRS of model object not set')

        if not self.inp.coordinates.empty:
            self.inp.coordinates = spatial.change_crs(self.inp.coordinates, self.crs, *args, **kwargs)

        if not self.inp.vertices.empty:
            self.inp.vertices = spatial.change_crs(self.inp.vertices, self.crs, *args, **kwargs)

        if not self.inp.polygons.empty:
            self.inp.polygons = spatial.change_crs(self.inp.polygons, self.crs, *args, **kwargs)

        self.crs = args[0]

    def to_geojson(self, target_path=None):
        """
        Return a GeoJSON representation of the entire model
        :param target_path: target path of geojson (optional)
        :return: GeoJSON representation of model
        """
        raise NotImplementedError

    def export_to_shapefile(self, shpdir, prj=None):
        """
        export the model data into a shapefile. element_type dictates which type
        of data will be included.

        default projection is PA State Plane - untested on other cases
        """

        # CREATE THE CONDUIT shp
        conds = self.conduits()
        conds_path = os.path.join(shpdir, self.inp.name + '_conduits.shp')
        spatial.write_shapefile(conds, conds_path, prj=prj)

        # CREATE THE NODE shp
        nodes = self.nodes()
        nodes_path = os.path.join(shpdir, self.inp.name + '_nodes.shp')
        spatial.write_shapefile(nodes, nodes_path, geomtype='point', prj=prj)


class SWMMIOFile(object):
    defaultSection = "Link Flow Summary"

    def __init__(self, file_path):

        # file name and path variables
        self.path = file_path
        self.name = os.path.splitext(os.path.basename(file_path))[0]
        self.dir = os.path.dirname(file_path)
        self.file_size = os.path.getsize(file_path)

    def findByteRangeOfSection(self, startStr):
        '''
        returns the start and end "byte" location of substrings in a text file
        '''

        with open(self.path) as f:
            start = None
            end = None
            l = 0  # line bytes index
            for line in f:

                if start and line.strip() == "" and (l - start) > 100:
                    # LOGIC: if start exists (was found) and the current line
                    # length is 3 or less (length of /n ) and we're more than
                    # 100 bytes from the start location then we are at the first
                    # "blank" line after our start section (aka the end of the
                    # section)
                    end = l
                    break

                if (startStr in line) and (not start):
                    start = l

                # increment length (bytes?) of current position
                l += len(line) + len("\n")

        return [start, end]


class rpt(SWMMIOFile):
    '''
    An accessible SWMM .rpt object
    '''

    def __init__(self, filePath):

        SWMMIOFile.__init__(self, filePath)

        with open(filePath) as f:
            for line in f:
                if "Starting Date" in line:
                    simulationStart = line.split(".. ")[1].replace("\n", "")
                if "Ending Date" in line:
                    simulationEnd = line.split(".. ")[1].replace("\n", "")
                if "Report Time Step ........." in line:
                    timeStepMin = int(line.split(":")[1].replace("\n", ""))
                    break

        self.simulationStart = simulationStart
        self.simulationEnd = simulationEnd
        self.timeStepMin = timeStepMin

        # grab the date of analysis
        with open(filePath) as f:
            f.seek(self.file_size - 500)  # jump to 500 bytes before the end of file
            for line in f:
                if "Analysis begun on" in line:
                    date = line.split("Analysis begun on:  ")[1].replace("\n", "")

        self.dateOfAnalysis = date
        self.elementByteLocations = {"Link Results": {}, "Node Results": {}}

    def returnDataAtDTime(self, id, dtime, sectionTitle="Link Results", startByte=0):
        '''
        return data from time series in RPT file
        '''

        byteLocDict = self.elementByteLocations[sectionTitle]
        if byteLocDict:
            startByte = byteLocDict[id]

        elif startByte == 0:
            startByte = self.findByteRangeOfSection(sectionTitle)[0]
            print('startByte ' + str(startByte))

        with open(self.path) as f:

            f.seek(startByte)  # jump to general area of file if we know it
            subsectionFound = False

            for line in f:
                if id in line: subsectionFound = True

                if subsectionFound and dtime in line:
                    line = ' '.join(re.findall('\"[^\"]*\"|\S+', line))
                    rowdata = line.replace("\n", "").split(" ")
                    return rowdata


class inp(SWMMIOFile):

    # creates an accessible SWMM .inp object
    # make sure INP has been saved in the GUI before using this

    def __init__(self, file_path):
        self._conduits_df = None
        self._junctions_df = None
        self._outfalls_df = None
        self._coordinates_df = None
        self._vertices_df = None
        self._polygons_df = None

        SWMMIOFile.__init__(self, file_path)  # run the superclass init

        self._sections = [self._conduits_df, self._junctions_df, self._outfalls_df,
                          self._coordinates_df, self._vertices_df, self._polygons_df]

    def save(self, target_path=None):
        '''
        Save the inp file to disk. File will be overwritten unless a target_path
        is provided
        '''
        from swmmio.utils.modify_model import replace_inp_section
        target_path = target_path if target_path is not None else self.path

        for section in self._sections:
            if section is not None:
                replace_inp_section()

    @property
    def conduits(self):
        """
        Get/set conduits section of the INP file.

        :return: Conduits section of the INP file
        :rtype: pandas.DataFrame

        Examples:

        >>> import swmmio
        >>> from swmmio.tests.data import MODEL_FULL_FEATURES__NET_PATH
        >>> model = swmmio.Model(MODEL_FULL_FEATURES__NET_PATH)
        >>> model.inp.conduits[['InletNode', 'OutletNode', 'Length', 'ManningN']]
              InletNode OutletNode  Length  ManningN
        Name
        C1:C2        J1         J2  244.63      0.01
        C2.1         J2         J3  666.00      0.01
        1             1          4  400.00      0.01
        2             4          5  400.00      0.01
        3             5         J1  400.00      0.01
        4             3          4  400.00      0.01
        5             2          5  400.00      0.01
        """
        if self._conduits_df is None:
            self._conduits_df = create_dataframeINP(self.path, "[CONDUITS]", comment_cols=False)
        return self._conduits_df

    @conduits.setter
    def conduits(self, df):
        """Set inp.conduits DataFrame."""
        self._conduits_df = df

    @property
    def junctions(self):
        """
        Get/set junctions section of the INP file.

        :return: junctions section of the INP file
        :rtype: pandas.DataFrame

        Examples:

        >>> import swmmio
        >>> from swmmio.tests.data import MODEL_FULL_FEATURES__NET_PATH
        >>> model = swmmio.Model(MODEL_FULL_FEATURES__NET_PATH)
        >>> model.inp.junctions
              InvertElev  MaxDepth  InitDepth  SurchargeDepth  PondedArea
        Name
        J3         6.547        15          0               0           0
        1         17.000         0          0               0           0
        2         17.000         0          0               0           0
        3         16.500         0          0               0           0
        4         16.000         0          0               0           0
        5         15.000         0          0               0           0
        J2        13.000        15          0               0           0
        """
        if self._junctions_df is None:
            self._junctions_df = create_dataframeINP(self.path, "[JUNCTIONS]", comment_cols=False)
        return self._junctions_df

    @junctions.setter
    def junctions(self, df):
        """Set inp.junctions DataFrame."""
        self._junctions_df = df

    @property
    def outfalls(self):
        """
        Get/set outfalls section of the INP file.

        :return: outfalls section of the INP file
        :rtype: pandas.DataFrame

        Examples:

        >>> import swmmio
        >>> from swmmio.tests.data import MODEL_FULL_FEATURES__NET_PATH
        >>> model = swmmio.Model(MODEL_FULL_FEATURES__NET_PATH)
        >>> model.inp.outfalls
              InvertElev OutfallType StageOrTimeseries  TideGate
        Name
        J4             0        FREE                NO       NaN
        """
        if self._outfalls_df is None:
            self._outfalls_df = create_dataframeINP(self.path, "[OUTFALLS]", comment_cols=False)
        return self._outfalls_df

    @outfalls.setter
    def outfalls(self, df):
        """Set inp.outfalls DataFrame."""
        self._outfalls_df = df

    @property
    def coordinates(self):
        """
        Get/set coordinates section of model
        :return: dataframe of model coordinates
        """
        if self._coordinates_df is not None:
            return self._coordinates_df
        self._coordinates_df = create_dataframeINP(self.path, "[COORDINATES]", comment_cols=False)
        return self._coordinates_df

    @coordinates.setter
    def coordinates(self, df):
        """Set inp.coordinates DataFrame."""
        self._coordinates_df = df

    @property
    def vertices(self):
        """
        get/set vertices section of model
        :return: dataframe of model coordinates
        """
        if self._vertices_df is not None:
            return self._vertices_df
        self._vertices_df = create_dataframeINP(self.path, '[VERTICES]', comment_cols=False)
        return self._vertices_df

    @vertices.setter
    def vertices(self, df):
        """Set inp.vertices DataFrame."""
        self._vertices_df = df

    @property
    def polygons(self):
        """
        get/set polygons section of model
        :return: dataframe of model coordinates
        """
        if self._polygons_df is not None:
            return self._polygons_df
        self._polygons_df = create_dataframeINP(self.path, '[Polygons]', comment_cols=False)
        return self._polygons_df

    @polygons.setter
    def polygons(self, df):
        """Set inp.polygons DataFrame."""
        self._polygons_df = df



#READ/WRITE REPORTS AS JSON
import json
import pandas as pd
from pandas.io.json import json_normalize
from swmmio.utils import spatial
from swmmio.graphics import swmm_graphics as sg
from swmmio.reporting.reporting import FloodReport
from swmmio.defs.config import *
import geojson


def decode_report(rpt_path):
    #read report from json into a dict
    with open(rpt_path, 'r') as f:
        read_rpt = json.loads(f.read())

    #parse the geojson
    def df_clean(uncleandf):
        cleaned_cols = [x.split('.')[-1] for x in uncleandf.columns]
        uncleandf.columns = cleaned_cols
        clean_df = uncleandf.rename(columns={'coordinates':'coords'}).drop(['type'], axis=1)
        clean_df = clean_df.set_index(['Name'])
        return clean_df

    #parse conduit data into a dataframe
    conds_df = json_normalize(read_rpt['conduits']['features'])
    conds_df = df_clean(conds_df)

    #parse node data into a dataframe
    nodes_df = json_normalize(read_rpt['nodes']['features'])
    nodes_df = df_clean(nodes_df)

    #parse parcel data into a dataframe
    pars_df = json_normalize(read_rpt['parcels']['features'])
    pars_df = df_clean(pars_df)

    rpt_dict = {'conduits':conds_df, 'nodes':nodes_df, 'parcels':pars_df}
    rpt_dict.update()
    return {'conduits':conds_df, 'nodes':nodes_df, 'parcels':pars_df}

def encode_report(rpt, rpt_path):

    rpt_dict = {}

    #write parcel json files
    parcels = spatial.read_shapefile(sg.config.parcels_shapefile)
    parcels = parcels[['PARCELID', 'coords']] #omit 'ADDRESS', 'OWNER1'
    flooded = rpt.alt_report.parcel_flooding #proposed flooding condition
    flooded = pd.merge(flooded, parcels, right_on='PARCELID', left_index=True)
    rpt_dict['parcels'] = spatial.write_geojson(flooded, geomtype='polygon')

    #non null delta category parcels
    delta_parcels = rpt.flood_comparison.loc[pd.notnull(rpt.flood_comparison.Category)]
    delta_parcels = pd.merge(delta_parcels, parcels, right_on='PARCELID', left_index=True)
    rpt_dict['delta_parcels'] = spatial.write_geojson(delta_parcels, geomtype='polygon')

    #encode conduit and nodes data into geojson
    # rpt_dict['conduits'] = spatial.write_geojson(rpt.alt_report.model.conduits())
    rpt_dict['new_conduits'] = spatial.write_geojson(rpt.newconduits)
    # rpt_dict['nodes'] = spatial.write_geojson(rpt.model.nodes(), geomtype='point')

    #write summary stats
    rpt_dict.update(rpt.summary_dict)

    with open(rpt_path, 'w') as f:
        f.write(json.dumps(rpt_dict))

    # #start writing that thing
    # with open(BETTER_BASEMAP_PATH, 'r') as bm:
    #     filename = os.path.join(os.path.dirname(rpt_path), rpt.model.name + '.html')
    #     with open(filename, 'wb') as newmap:
    #         for line in bm:
    #             if '//INSERT GEOJSON HERE ~~~~~' in line:
    #                 newmap.write('conduits = {};\n'.format(geojson.dumps(rpt_dict['conduits'])))
    #                 newmap.write('nodes = {};\n'.format(geojson.dumps(rpt_dict['nodes'])))
    #                 newmap.write('parcels = {};\n'.format(geojson.dumps(rpt_dict['parcels'])))
    #             else:
    #                 newmap.write(line)



#UTILITY/HELPER FUNCTOINS FOR DRAWING
import pandas as pd
import math
from swmmio.graphics import * #constants
import os
from PIL import Image

def save_image(img, img_path, model=None, imgName=None, imgDir=None,
				antialias=True, open=False, fileExt=".png"):

	#get the size from the Image object
	imgSize = (img.getbbox()[2], img.getbbox()[3])
	if antialias:
		size = (int(imgSize[0]*0.5), int(imgSize[1]*0.5))
		img.thumbnail(size, Image.ANTIALIAS)

	img.save(img_path)
	if open:
		os.startfile(img_path)

def px_to_irl_coords(df, px_width=4096.0, bbox=None, shift_ratio=None):
	"""
	given a dataframe with element id (as index) and X1, Y1 columns (and
	optionally X2, Y2 columns), return a dataframe with the coords as pixel
	locations based on the targetImgW.
	"""

	df = df.loc[pd.notnull(df.coords)]

	if not bbox:
		xs = [xy[0] for verts in df.coords.tolist() for xy in verts]
		ys = [xy[1] for verts in df.coords.tolist() for xy in verts]
		xmin, ymin, xmax, ymax = (min(xs), min(ys), max(xs), max(ys))
		bbox = [(xmin, ymin),(xmax, ymax)]

	else:
		df = clip_to_box(df, bbox) #clip if necessary
		xmin = float(bbox[0][0])
		ymin = float(bbox[0][1])

	#find the actual dimensions, use to find scale factor
	height = bbox[1][1] - bbox[0][1]
	width = bbox[1][0] - bbox[0][0]

	if not shift_ratio:
		#to scale down from coordinate to pixels
		shift_ratio = float(px_width / width)

	def shft_coords(row):
		#parse through coords (nodes, or link) and adjust for pixel space
		return [(int((xy[0] - xmin)*shift_ratio),
					int((height - xy[1] + ymin)*shift_ratio))
						for xy in row.coords]

	#insert new column with the shifted coordinates
	draw_coords = df.apply(lambda row:shft_coords(row), axis=1)
	if not (draw_coords.empty and df.empty):
		df = df.assign(draw_coords = draw_coords)

	return df, bbox, int(height*shift_ratio),int(width*shift_ratio),shift_ratio

def circle_bbox(coordinates, radius=5):
    """the bounding box of a circle given as centriod coordinate and radius"""

    x = coordinates[0]
    y = coordinates[1]
    r = radius

    return (x-r, y-r, x+r, y+r)

def clip_to_box(df, bbox):
	"""clip a dataframe with a coords column to a bounding box"""
	def any_xy_in_box(row, bbox):
		#because im confused with list comprehensions rn
	    return any([point_in_box(bbox, pt) for pt in row])

	coords = df.coords.tolist()
	result = [any_xy_in_box(p, bbox) for p in coords]
	return df.loc[result]

def angle_bw_points(xy1, xy2):
	dx, dy = (xy2[0] - xy1[0]), (xy2[1] - xy1[1])

	angle = (math.atan(float(dx)/float(dy)) * 180/math.pi )
	if angle < 0:
		angle = 270 - angle
	else:
		angle = 90 - angle
	#angle in radians
	return angle

def midpoint(xy1, xy2):

	dx, dy = (xy2[0] + xy1[0]), (xy2[1] + xy1[1])
	midpt = ( int(dx/2), int(dy/2.0) )

	#angle in radians
	return midpt
def point_in_box(bbox, point):
	"""check if a point falls with in a bounding box, bbox"""
	LB = bbox[0]
	RU = bbox[1]

	x = point[0]
	y = point[1]

	if x < LB[0] or x >  RU[0]:
		return False
	elif y < LB[1] or y > RU[1]:
		return False
	else:
		return True

def length_bw_coords(upstreamXY, downstreamXY):

    #return the distance (units based on input) between two points
    x1 = float(upstreamXY[0])
    x2 = float(downstreamXY[0])
    y1 = float(upstreamXY[1])
    y2 = float(downstreamXY[1])

    return math.hypot(x2 - x1, y2 - y1)



from swmmio.run_models import run
from swmmio.swmmio import Model
# from swmmio.reporting import reporting
# from swmmio.utils import swmm_utils as su
from swmmio.reporting import batch
from multiprocessing import Pool, cpu_count
from datetime import datetime
import os
import sys
wd = os.getcwd()

log_start_time = datetime.now().strftime("%y%m%d_%H%M")
def run_swmm_engine(inp_folder):

    logfile = os.path.join(wd, 'log_'+log_start_time+'.txt')

    m = Model(inp_folder)
    if not m.rpt_is_valid():
        # if the rpt is not valid i.e. not having current, usable data: run
        with open (logfile, 'a') as f:
            now = datetime.now().strftime("%y-%m-%d %H:%M")
            f.write('{}: started at {} '.format(m.inp.name, now))
            # print 'running {}\n'.format(m.inp.name)
            run.run_hot_start_sequence(m.inp.path)
            now = datetime.now().strftime("%y-%m-%d %H:%M")
            f.write(', completed at {}\n'.format(now))
    else:
        with open (logfile, 'a') as f:
            f.write('{}: skipped (up-to-date)\n'.format(m.inp.name))

def main(inp_paths, cores_left):

    """
    called from the cli:
    swmmio -sp DIR1, DIR2, ... -cores_left=4
    """

    # create multiprocessing Pool object using all cores less the -cores_left
    #beware of setting -cores_left=0, CPU usage will max the F out
    pool = Pool(cpu_count() - cores_left)

    #create a process pool with the run_swmm_engine() func on each directory
    res = pool.map(run_swmm_engine, inp_paths)

    # #post process
    # ADMIN_DIR = os.path.join(wd, 'ProjectAdmin')
    # results_file = os.path(ADMIN_DIR, '170210_results.csv')
    # add_data = os.path(ADMIN_DIR, 'additional_costs.csv')
    # join_data = os.path(ADMIN_DIR, 'equivalentgeoms.csv')
    # batch.batch_reports(wd, results_file=results_file,
    #                     additional_costs=add_data, join_data=join_data)



"""
Objects encapsulating model elements
"""
import swmmio
from swmmio.utils.dataframes import create_dataframeINP, create_dataframeRPT, get_link_coords
from swmmio.tests.data import MODEL_FULL_FEATURES__NET_PATH
from swmmio.defs import HEADERS
from swmmio.utils.spatial import coords_series_to_geometry


class ModelSection(object):
    def __init__(self, model, section_name):
        """
        Base class of a group of model elements.
        :param model: swmmio.Model object
        :param section_name: name of section of model
        """
        self.model = model
        self.inp = self.model.inp
        self.rpt = self.model.rpt
        self.section_name = section_name
        self.config = HEADERS[section_name.lower()]

    # def to_geojson(self, target_path=None):
    #     """
    #     Return a GeoJSON representation of the group
    #     :param target_path: target of GeoJSON representation of elements
    #     :return: GeoJSON representation of elements
    #     """
    def to_gdf(self):
        # uses GeoPandas
        try:
            import geopandas as gp
        except ImportError:
            raise ImportError('geopandas module needed. Install GeoPandas with conda: ',
                              'conda install geopandas')

        df = self.__call__()
        df['geometry'] = coords_series_to_geometry(df['coords'], geomtype='linestring', format='shape')
        df = df.drop(['coords'], axis=1)
        return gp.GeoDataFrame(df, crs=self.model.crs)

    def __call__(self):

        """
        collect all useful and available data related to the conduits and
        organize in one dataframe.
        >>> model = swmmio.Model(MODEL_FULL_FEATURES__NET_PATH)
        >>> conduits_section = ModelSection(model, 'conduits')
        >>> conduits_section()
        """

        # create dataframes of relevant sections from the INP
        for ix, sect in enumerate(self.config['inp_sections']):
            if ix == 0:
                df = create_dataframeINP(self.inp.path, sect, comment_cols=False)
            else:
                df_other = create_dataframeINP(self.inp.path, sect, comment_cols=False)
                df = df.join(df_other)

        if self.rpt:
            for rpt_sect in self.config['rpt_sections']:
                df = df.join(create_dataframeRPT(self.rpt.path, rpt_sect))

        # add conduit coordinates
        xys = df.apply(lambda r: get_link_coords(r, self.inp.coordinates, self.inp.vertices), axis=1)
        df = df.assign(coords=xys.map(lambda x: x[0]))

        # make inlet/outlet node IDs string type
        df.InletNode = df.InletNode.astype(str)
        df.OutletNode = df.OutletNode.astype(str)

        return df




#Module is intended to provide high level access to post processing routines
#such that standard reporting and figures can be generated to report on the
#perfomance of given SFR alternatives/options
from swmmio.reporting.functions import *
from swmmio.reporting.utils import insert_in_file, insert_in_file_2
from swmmio.damage import parcels
from swmmio.graphics import swmm_graphics as sg
from swmmio.graphics import drawing
from swmmio.graphics.constants import *
from swmmio.utils.dataframes import create_dataframeRPT
from swmmio.utils import spatial
from swmmio.version_control.inp import INPDiff
from swmmio import swmmio
import os
import math
import pandas as pd
from swmmio.defs.config import *
import json, geojson
import shutil


class FloodReport(object):
    def __init__(self, model=None, parcel_node_df=None, parcel_node_join_csv=None,
                 threshold=0.0833):
        """
        Report of parcel flood duration of a given swmmio.Model object
        """
        self.total_parcel_count = None
        self.model = None
        self.scenario = None
        self.parcel_flooding = None
        self.cost_estimate = None
        self.new_conduits_geojson = None
        self.parcel_hrs_flooded = None
        self.parcel_vol_flooded = None

        if model is not None:
            if parcel_node_df is None:
                #read csv if df not provided
                parcel_node_df = pd.read_csv(parcel_node_join_csv)

            self.total_parcel_count = len(parcel_node_df.PARCELID.unique())
            self.model = model
            self.scenario = model.scenario
            self.parcel_flooding = parcels.flood_duration(model.nodes(),
                                                          parcel_node_df,
                                                          threshold=threshold)
            self.parcel_hrs_flooded = self.parcel_flooding.HoursFlooded.sum()
            subs = self.model.subcatchments()
            nodes = self.model.nodes()
            self.runoff_vol_mg = subs.RunoffMGAccurate.sum()
            self.flood_vol_mg = nodes.TotalFloodVol.sum()

    def __str__(self):
        """print friendly"""
        a = '{}\n'.format(self.model.name)
        b = self.duration_partition()
        c = '\nRunoff: {}MG'.format(round(self.runoff_vol_mg, 1))

        ffrac = round(self.flood_vol_mg / self.runoff_vol_mg *100, 1)
        d = 'Volume Flooded: {}MG ({}%)'.format(round(self.flood_vol_mg, 1), ffrac)

        #return '{}\n{}'.format(a, '\n'.join(b))
        return '\n'.join([a]+ b + [c, d])


    def duration_partition(self, partitions=[5, 10, 15, 30, 60, 120], raw=False):
        pflood = self.parcel_flooding
        pcount = self.total_parcel_count
        results = []
        for dur in partitions:
            n = len(pflood.loc[pflood.HoursFlooded > dur/60.0])
            frac = n/float(pcount)*100.0
            if raw:

                results.append(s)
            else:
                s = '{}mins: {} ({}%)'.format(dur, n, round(frac,1))
            results.append(s)
        return results


class ComparisonReport(object):
    def __init__(self, baseline_report, alt_report, additional_costs=None,
                 join_data=None):
        """
        Report object representing a comparison of a baseline report to another
        (proposed conditions) report
        """
        basemodel = baseline_report.model
        altmodel = alt_report.model
        baseline_flooding = baseline_report.parcel_flooding
        proposed_flooding = alt_report.parcel_flooding

        conduitdiff = INPDiff(basemodel, altmodel, '[CONDUITS]')
        new_cond_ids = pd.concat([conduitdiff.added, conduitdiff.altered]).index

        self.baseline_report = baseline_report
        self.alt_report = alt_report

        #sort out the new and "altered" conduits
        self.newconduits = altmodel.conduits().loc[new_cond_ids]
        self.newconduits.loc[conduitdiff.altered.index, 'Category'] = 'Replaced'
        self.newconduits.loc[conduitdiff.added.index, 'Category'] = 'Proposed'
        self.new_ix = self.newconduits.index

        #human readable name
        self.name = '{} vs {} Report'.format(basemodel.name, altmodel.name)
        phase_diff = list(set(altmodel.name.split('_')) - set(basemodel.name.split('_')))
        phase_diff = ', '.join(phase_diff[:-1]) + ', and {}'.format(phase_diff[-1])
        self.description = 'Adding {} to {}'.format(phase_diff, basemodel.name)

        #calculate the proposed sewer mileage
        proposed_ft = self.newconduits.Length.sum()#conduitdiff.added.Length.sum() + conduitdiff.altered.Length.sum()
        self.sewer_miles_new = proposed_ft / 5280.0

        #ADD ANY ADDITIONAL DATA TO THE NEW CONDIUTS (FEASIBILITY, REAL GEOM etc)
        if join_data is not None:
            joindf = pd.read_csv(join_data, index_col=0)
            self.newconduits = self.newconduits.join(joindf)

        #COST ESTIMATION
        self.newconduits = conduits_cost_estimate(self.newconduits, additional_costs)
        self.cost_estimate = self.newconduits.TotalCostEstimate.sum() / math.pow(10, 6)

        #MEASURE THE FLOOD REDUCTION IMPACT
        self.flood_comparison = parcels.compare_flood_duration(baseline_flooding,
                                                               proposed_flooding)
        self.impact = self.flood_comparison.Category.value_counts()

        df = self.flood_comparison
        parcel_hours_reduction = df.loc[df.Category.isin(['decreased_flooding',
                                                          'eliminated_flooding']),
                                        'DeltaHours']

        parcel_hours_increased = df.loc[df.Category == 'increased_flooding','DeltaHours']
        parcel_hours_new = df.loc[df.Category == 'new_flooding','DeltaHours']
        self.parcel_hours_reduced = - sum(parcel_hours_reduction)
        self.parcel_hours_increased = sum(parcel_hours_increased)
        self.parcel_hours_new = sum(parcel_hours_new)

        self.summary_dict = {
            'parcel_hours_reduced':self.parcel_hours_reduced,
            'parcel_hours_increased':self.parcel_hours_increased,
            'parcel_hours_new':self.parcel_hours_new,
            'description':self.description,
            'cost_estimate':self.cost_estimate,
            'baseline_name':self.baseline_report.model.name,
            'alt_name':self.alt_report.model.name,
            'sewer_miles_new':self.sewer_miles_new,
        }



    def write(self, rpt_dir):
        #write cost per sewer segment spreadsheet
        self.newconduits.to_csv(os.path.join(rpt_dir,'cost_estimate.csv'))
        self.flood_comparison.to_csv(os.path.join(rpt_dir,'parcel_flood_comparison.csv'))

        #write parcel json files
        parcels = spatial.read_shapefile(sg.config.parcels_shapefile)
        parcels = parcels[['PARCELID', 'ADDRESS', 'OWNER1', 'coords']]
        flooded = self.flood_comparison
        flooded = flooded.loc[flooded.Category.notnull()] #parcels with significant flood delta
        flooded = pd.merge(flooded, parcels, right_on='PARCELID', left_index=True)
        colors = flooded.apply(lambda row:'#%02x%02x%02x' % drawing.parcel_draw_color(row, style='delta'), axis=1)
        flooded = flooded.assign(fill=colors)
        geoparcelpath = os.path.join(rpt_dir,'delta_parcels.json')
        spatial.write_geojson(flooded, filename=geoparcelpath, geomtype='polygon')

        #write new conduit json, shapefiles
        shpdir = os.path.join(os.path.dirname(rpt_dir), 'shapefiles')
        if not os.path.exists(shpdir):os.mkdir(shpdir)
        geocondpath = os.path.join(rpt_dir,'new_conduits.json')
        shpcondpath = os.path.join(shpdir, self.alt_report.model.inp.name + '_new_conduits.shp')
        spatial.write_geojson(self.newconduits, filename=geocondpath)
        spatial.write_shapefile(self.newconduits, filename=shpcondpath)

        #write node and conduit report csvs
        self.alt_report.model.nodes().to_csv(os.path.join(rpt_dir,'nodes.csv'))
        self.alt_report.model.conduits().to_csv(os.path.join(rpt_dir,'conduits.csv'))

        #write a html map
        with open (geocondpath, 'r') as f:
            geo_conduits = geojson.loads(f.read())


        proposed_flooded = self.alt_report.parcel_flooding
        proposed_flooded = pd.merge(proposed_flooded, parcels, right_on='PARCELID', left_index=True)
        geo_parcels = spatial.write_geojson(proposed_flooded)
        # with open (geoparcelpath, 'r') as f:
        #     geo_parcels = geojson.loads(f.read())

        with open(BETTER_BASEMAP_PATH, 'r') as bm:
            filename = os.path.join(os.path.dirname(geocondpath), self.alt_report.model.name + '.html')
            with open(filename, 'wb') as newmap:
                for line in bm:
                    if '//INSERT GEOJSON HERE ~~~~~' in line:
                        newmap.write('conduits = {};\n'.format(geojson.dumps(geo_conduits)))
                        newmap.write('nodes = {};\n'.format(0))
                        newmap.write('parcels = {};\n'.format(geojson.dumps(geo_parcels)))
                    else:
                        newmap.write(line)


        #create figures
    def generate_figures(self, rpt_dir, parcel_shp_df, bbox=d68d70):

        basemodel = self.baseline_report.model
        altmodel = self.alt_report.model
        files = '{}\n{}'.format(basemodel.inp.path, altmodel.inp.path)


        #SIMPLE PLAN VIEW OF OPTION (showing new conduits)
        conduits = altmodel.conduits()
        pth = os.path.join(rpt_dir, '00 Proposed Infrastructure.png')
        conduits['draw_color'] = '#bebeb4' #default color
        conduits['draw_size'] = conduits.Geom1
        conduits.loc[self.new_ix, 'draw_color'] = '#1414e6' #new cond col
        conduits.loc[self.new_ix,'draw_size'] = conduits.loc[self.new_ix,'Geom1']*2
        sg.draw_model(conduits=conduits, nodes=altmodel.nodes(), bbox=bbox,
                      title=self.name, annotation=files, file_path=pth)

        #EXISTING CONDITIONS PARCEL FLOOD DURATION
        pth = os.path.join(rpt_dir, '01 Existing Parcel Flood Duration.png')
        base_conduits = basemodel.conduits()
        base_conduits['draw_color'] = '#bebeb4' #default color
        base_conduits['draw_size'] = base_conduits.Geom1
        pars = self.baseline_report.parcel_flooding
        pars = pd.merge(pars, parcel_shp_df, right_on='PARCELID', left_index=True)
        comp_cols = pars.apply(lambda row: drawing.parcel_draw_color(row, 'risk'), axis=1)
        pars = pars.assign(draw_color=comp_cols)
        sg.draw_model(conduits=base_conduits, nodes=basemodel.nodes(), parcels=pars,
                      bbox=bbox,title=self.name, annotation=self.baseline_report.__str__(),
                      file_path=pth)

        #PROPOSED CONDITIONS PARCEL FLOOD DURATION
        pth = os.path.join(rpt_dir, '02 Proposed Parcel Flood Duration.png')
        pars = self.alt_report.parcel_flooding
        pars = pd.merge(pars, parcel_shp_df, right_on='PARCELID', left_index=True)
        comp_cols = pars.apply(lambda row: drawing.parcel_draw_color(row, 'risk'), axis=1)
        pars = pars.assign(draw_color=comp_cols)
        sg.draw_model(conduits=conduits, nodes=altmodel.nodes(), parcels=pars,
                      bbox=bbox,title=self.name, annotation=self.alt_report.__str__(),
                      file_path=pth)

        #IMPACT OF INFRASTRUCTURE
        pth = os.path.join(rpt_dir, '03 Impact of Option.png')
        pars = pd.merge(self.flood_comparison, parcel_shp_df, right_on='PARCELID', left_index=True)
        comp_cols = pars.apply(lambda row: drawing.parcel_draw_color(row, 'delta'), axis=1)
        pars = pars.assign(draw_color=comp_cols)
        annotation = self.alt_report.__str__() + '\n\n' + self.__str__()
        sg.draw_model(conduits=conduits, nodes=altmodel.nodes(), parcels=pars,
                      bbox=bbox,title=self.name, annotation=annotation,
                      file_path=pth)

        #IMPACT OF INFRASTRUCTURE OVERALL
        pth = os.path.join(rpt_dir, '04 Impact of Option - Overall.png')
        sg.draw_model(conduits=conduits, nodes=altmodel.nodes(), parcels=pars,
                      bbox=study_area,title=self.name, annotation=annotation,
                      file_path=pth)

    def __str__(self):
        """print friendly"""
        catz = [_f for _f in self.flood_comparison.Category.unique() if _f]
        a = ['{}: {}'.format(c, self.impact[c]) for c in catz]
        files = [self.baseline_report.model.inp.path,
                 self.alt_report.model.inp.path]
        return '{}\n{}'.format(self.name,'\n'.join(a+files))


def read_report_dir(rptdir, total_parcel_count=0):
    #rpt dir passed in, just read the preprossed report data
    rpt = FloodReport()
    rpt.total_parcel_count = total_parcel_count
    rpt.model = swmmio.Model(os.path.dirname(rptdir))
    rpt.scenario = rpt.model.scenario
    rpt.parcel_flooding = pd.read_csv(os.path.join(rptdir,
                                                   'parcel_flood_comparison.csv'))
    rpt.parcel_hrs_flooded = rpt.parcel_flooding.HoursFloodedProposed.sum()
    rpt.parcel_vol_flooded = rpt.parcel_flooding.TotalFloodVolProposed.sum()
    costcsv = os.path.join(rptdir, 'cost_estimate.csv')
    conduits_geojson_path = os.path.join(rptdir, 'new_conduits.json')

    if os.path.exists(costcsv):
        #calc the cost estimate total in millions
        cost_df =  pd.read_csv(costcsv)
        rpt.cost_estimate = cost_df.TotalCostEstimate.sum() / math.pow(10, 6)

    if os.path.exists(conduits_geojson_path):
        with open (conduits_geojson_path, 'r') as f:
            rpt.new_conduits_geojson = geojson.loads(f.read())

    return rpt



import os, shutil
import pandas as pd
from swmmio.version_control.inp import INPDiff
from swmmio.defs.config import *


def create_map(model1, model2=None, bbox=None, crs=None, filename=None,
               subset=None, return_data=False):

    """
    export model as a geojson object. this is sloppy currently
    """

    import geojson
    from geojson import Point, LineString, Feature, GeometryCollection, FeatureCollection
    try: import pyproj
    except ImportError:
        raise ImportError('pyproj module needed. get this package here: ',
                        'https://pypi.python.org/pypi/pyproj')


    if model2 is not None:
        changes = INPDiff(model1, model2, section='[CONDUITS]')
        df = pd.concat([changes.added, changes.altered])
        subset = df.index.tolist()

    # else:
    #     model2 = model1 #stupid, for file path


    if crs is None:
        #default coordinate system (http://spatialreference.org/ref/epsg/2272/):
        #NAD_1983_StatePlane_Pennsylvania_South_FIPS_3702_Feet
        crs =  {"type": "name","properties": {"name": "EPSG:2272"}}

    pa_plane = pyproj.Proj(init='epsg:2272', preserve_units=True)
    wgs = pyproj.Proj(proj='latlong', datum='WGS84', ellps='WGS84') #google maps, etc


    geometries = [] #array of features
    #collect the links
    for k,v in list(model2.list_objects('conduit', bbox, subset=subset).items()):
        props = {
            'MaxQPercent':v.maxQpercent,
            'id':v.id,
            'lifecycle':v.lifecycle,
            'geom1':v.geom1,
            'geom2':v.geom2,
            }

        latlngs = [pyproj.transform(pa_plane, wgs, *xy) for xy in v.coordinates]
        geometry = LineString(latlngs, properties=props)
        geometries.append(geometry)


    if return_data is True:
        return FeatureCollection(geometries, crs=crs)

    if filename is None:
        filename = os.path.join(model2.inp.dir,
                                REPORT_DIR_NAME,
                                model2.name + '.html')

    #start writing that thing
    with open(BASEMAP_PATH, 'r') as bm:
        with open(filename, 'wb') as newmap:
            for line in bm:
                if 'SWMMIO DUMP JSON DATA HERE' in line:
                    newmap.write('data = ')
                    newmap.write(geojson.dumps(FeatureCollection(geometries, crs=crs)))
                else:
                    newmap.write(line)



import pandas as pd
import os
from os.path import join as j
from swmmio.defs.sectionheaders import inp_header_dict

current_dir = os.path.dirname(__file__)
sets = j(current_dir, 'inp_settings')


OPTIONS_no_rain = pd.read_table(
                                j(sets, 'OPTIONS_no_rain.txt'),
                                delim_whitespace=True,
                                names=inp_header_dict['[OPTIONS]'].split(),
                                skiprows=1,
                                index_col=0
                                )

OPTIONS_normal = pd.read_table(
                                j(sets, 'OPTIONS_normal.txt'),
                                delim_whitespace=True,
                                names=inp_header_dict['[OPTIONS]'].split(),
                                skiprows=1,
                                index_col=0
                                )
REPORT_nodes_links = pd.read_table(
                                j(sets, 'REPORT_nodes_links.txt'),
                                delim_whitespace=True,
                                names=inp_header_dict['[REPORT]'].split(),
                                skiprows=1,
                                index_col=0
                                )
REPORT_none = pd.read_table(
                                j(sets, 'REPORT_none.txt'),
                                delim_whitespace=True,
                                names=inp_header_dict['[REPORT]'].split(),
                                skiprows=1,
                                index_col=0
                            )



#graphical functions for SWMM files
from swmmio.defs.config import *
from swmmio.damage import parcels as pdamage
from swmmio.graphics import config, options
from swmmio.graphics.constants import * #constants
from swmmio.graphics.utils import *
from swmmio.graphics.drawing import *
from swmmio.utils import spatial
import pandas as pd
import os
from PIL import Image, ImageDraw



def _draw_basemap(draw, img, bbox, px_width, shift_ratio):

	"""
	given the shapefiles in options.basemap_options, render each layer
	on the model basemap.
	"""

	for f in options.basemap_options['features']:

		shp_path = os.path.join(config.basemap_shapefile_dir, f['feature'])
		df = spatial.read_shapefile(shp_path)[f['cols']+['coords']]
		df = px_to_irl_coords(df, bbox=bbox, shift_ratio=shift_ratio,
								px_width=px_width)[0]

		if 'ST_NAME' in df.columns:
			#this is a street, draw a polyline accordingly
			df.apply(lambda r: draw.line(r.draw_coords, fill=f['fill']), axis=1)
			annotate_streets(df, img, 'ST_NAME')
		else:
			df.apply(lambda r: draw.polygon(r.draw_coords,
											fill=f['fill']), axis=1)


def draw_model(model=None, nodes=None, conduits=None, parcels=None, title=None,
				annotation=None, file_path=None, bbox=None, px_width=2048.0):
	"""
	create a png rendering of the model and model results.

	A swmmio.Model object can be passed in independently, or Pandas Dataframes
	for the nodes and conduits of a model may be passed in. A dataframe containing
	parcel data can optionally be passed in.

	model -> swmmio.Model object

	nodes -> Pandas Dataframe (optional, if model not provided)

	conduits -> Pandas Dataframe (optional, if model not provided)

	parcels - > Pandas Dataframe (optional)

	title -> string, to be written in top left of PNG

	annotation -> string, to be written in bottom left of PNG

	file_path -> stirng, file path where png should be drawn. if not specified,
		a PIL Image object is return (nice for IPython notebooks)

	bbox -> tuple of coordinates representing bottom left and top right corner
		of a bounding box. the rendering will be clipped to this box. If not
		provided, the rendering will clip tightly to the model extents
		e.g. bbox = ((2691647, 221073),	(2702592, 227171))

		Note: this hasn't been tested with anything other than PA StatePlane coords

	px_width -> float, width of image in pixels
	"""

	#gather the nodes and conduits data if a swmmio Model object was passed in
	if model is not None:
		nodes = model.nodes()
		conduits = model.conduits()

	#antialias X2
	xplier=1
	xplier *= px_width/1024 #scale the symbology sizes
	px_width = px_width*2

	#compute draw coordinates, and the image dimensions (in px)
	conduits, bb, h, w, shift_ratio = px_to_irl_coords(conduits, bbox=bbox, px_width=px_width)
	nodes = px_to_irl_coords(nodes, bbox=bb, px_width=px_width)[0]

	#create the PIL image and draw objects
	img = Image.new('RGB', (w,h), white)
	draw = ImageDraw.Draw(img)

	#draw the basemap if required
	if config.include_basemap is True:
		_draw_basemap(draw, img, bb, px_width, shift_ratio)

	if parcels is not None:
		#expects dataframe with coords and draw color column
		par_px = px_to_irl_coords(parcels, bbox=bb, shift_ratio=shift_ratio, px_width=px_width)[0]
		par_px.apply(lambda r: draw.polygon(r.draw_coords, fill=r.draw_color), axis=1)

	# if config.include_parcels is True:
	# 	par_flood = pdamage.flood_duration(nodes, parcel_node_join_csv=config.parcel_node_join_data)
	# 	par_shp = spatial.read_shapefile(config.parcels_shapefile)
	# 	par_px = px_to_irl_coords(par_shp, bbox=bb, shift_ratio=shift_ratio, px_width=px_width)[0]
	# 	parcels = pd.merge(par_flood, par_px, right_on='PARCELID', left_index=True)
	# 	parcels.apply(lambda row: draw_parcel_risk(row, draw), axis=1)

	#start the draw fest, mapping draw methods to each row in the dataframes
	conduits.apply(lambda row: draw_conduit(row, draw), axis=1)
	nodes.apply(lambda row: draw_node(row, draw), axis=1)

	#ADD ANNOTATION AS NECESSARY
	if title: annotate_title(title, draw)
	if annotation: annotate_details(annotation, draw)
	annotate_timestamp(draw)

	#SAVE IMAGE TO DISK
	if file_path:
		save_image(img, file_path)

	return img



#METHODS USED TO VALIDATE INP FILES (e.g. ensure no duplicates exists)

from swmmio.utils import functions as funcs
from swmmio.utils import dataframes

def search_for_duplicates(inp_path, verbose = False):
    """
    scan an inp file and determine if any element IDs are duplicated in
    any section. Method: count the uniques and compare to total length
    """
    headers = funcs.complete_inp_headers(inp_path)['headers']
    dups_found = False
    for header, cols, in headers.items():
        if cols != 'blob':

            df = dataframes.create_dataframeINP(inp_path, section=header)
            elements = df.index
            n_unique = len(elements.unique()) #number of unique elements
            n_total = len(elements) #total number of elements
            if verbose:
                print('{} -> (uniques, total) -> ({}, {})'.format(header, n_unique , n_total))

            if n_unique != n_total:
                dups = ', '.join(df[df.index.duplicated()].index.unique().tolist())
                print('duplicate found in {}\nsection: {}\n{}'.format(inp_path, header, dups))
                dups_found = True

    return dups_found



from .run_models.run import run_simple, run_hot_start_sequence
from .run_models import start_pool
from .swmmio import Model
from itertools import chain
import os
import argparse
from multiprocessing import Pool, cpu_count
from datetime import datetime

#parse the arguments
parser = argparse.ArgumentParser(description='Process some stuff')
parser.add_argument('-r', '--run', dest='model_to_run', nargs="+")
parser.add_argument('-rhs', '--run_hotstart', dest='hotstart_model_to_run', nargs="+")
parser.add_argument('-sp', '--start_pool', dest='start_pool', nargs="+")
parser.add_argument('-cores_left', '--cores_left', dest='cores_left', default=4, type=int)
parser.add_argument('-pp', '--post_process', dest='post_process', nargs="+")

args = parser.parse_args()
wd = os.getcwd() #current directory script is being called from

if args.model_to_run is not None:

    models_paths = [os.path.join(wd, f) for f in args.model_to_run]
    print('Adding models to queue:\n\t{}'.format('\n\t'.join(models_paths)))

    #run the models in series (one after the other)
    list(map(run_simple, models_paths))
    # run_simple(args.model_to_run)

elif args.hotstart_model_to_run is not None:
    models_paths = [os.path.join(wd, f) for f in args.hotstart_model_to_run]
    print('hotstart_model_to_run the model: {}'.format(args.hotstart_model_to_run))
    # m = Model(args.hotstart_model_to_run)
    # run_hot_start_sequence(m)#args.hotstart_model_to_run)
    list(map(run_hot_start_sequence, models_paths))

elif args.start_pool is not None:

    models_dirs = [os.path.join(wd, f) for f in args.start_pool]
    print('Searching for models in:\n\t{}'.format('\n\t'.join(models_dirs)))
    #combine the segments and options (combinations) into one iterable
    inp_paths = []
    for root, dirs, files in chain.from_iterable(os.walk(path) for path in models_dirs):
        for f in files:
            if f.endswith('.inp') and 'bk' not in root:
                #we've found a directory containing an inp
                inp_paths.append(os.path.join(root, f))


    #call the main() function in start_pool.py
    start_pool.main(inp_paths, args.cores_left)

    print("swmmio has completed running {} models".format(len(inp_paths)))


else:
    print('you need to pass in some args')



#=================
#DEFINE INP HEADER THAT SHOULD BE REPLACED
#=================


inp_header_dict = {
    '[TITLE]':'blob',
    '[OPTIONS]':'Name Value',
    '[CONDUITS]': 'Name InletNode OutletNode Length ManningN InletOffset OutletOffset InitFlow MaxFlow',
    '[COORDINATES]': 'Name X Y',
    '[JUNCTIONS]': 'Name InvertElev MaxDepth InitDepth SurchargeDepth PondedArea',
    '[ORIFICES]': 'Name InletNode OutletNode OrificeType CrestHeight DischCoeff FlapGate OpenCloseTime',
    '[OUTFALLS]': 'Name InvertElev OutfallType StageOrTimeseries TideGate',
    '[STORAGE]': 'Name InvertElev MaxD InitDepth StorageCurve Coefficient Exponent Constant PondedArea EvapFrac SuctionHead Conductivity InitialDeficit',
    '[VERTICES]': 'Name X Y',
    '[WEIRS]': 'Name InletNode OutletNode WeirType CrestHeight DischCoeff FlapGate EndCon EndCoeff',
    '[PUMPS]':'Name',
    '[XSECTIONS]':'Link Shape Geom1 Geom2 Geom3 Geom4 Barrels',
    '[SUBCATCHMENTS]':'Name Raingage Outlet Area PercImperv Width PercSlope CurbLength SnowPack',
    '[SUBAREAS]':'Name N-Imperv N-Perv S-Imperv S-Perv PctZero RouteTo PctRouted',
    '[LOSSES]':'Link Inlet Outlet Average FlapGate',
    '[PUMPS]':'Name InletNode OutletNode PumpCurve InitStatus Depth ShutoffDepth',
    '[DWF]':'Node Parameter AverageValue TimePatterns',
    '[RAINGAGES]':'Name RainType TimeIntrv SnowCatch DataSourceType DataSourceName',
    '[INFILTRATION]':'Subcatchment Suction HydCon IMDmax',
    '[Polygons]':'Name X Y',
    '[REPORT]':'Param Status',
    '[TAGS]':'ElementType Name Tag',
    '[MAP]': 'x1 y1 x2 y2'
    #'[CURVES]':'Name Type X-Value Y-Value',
    #'[TIMESERIES]':'Name Date Time Value'
}

#=================
#DEFINE RPT HEADER THAT SHOULD BE REPLACED
#=================
rpt_header_dict={
    'Subcatchment Summary':"Name Area Width ImpervPerc SlopePerc RainGage Outlet",
    'Node Summary':"Name Type InvertEl MaxD PondedA ExternalInf",
    'Link Summary':'Name FromNode ToNode Type Length SlopePerc Roughness',
    'Cross Section Summary':'Name Shape DFull AreaFull HydRad MaxW NumBarrels FullFlow',
    'Subcatchment Runoff Summary':'Name TotalPrecip TotalRunon TotalEvap TotalInfil TotalRunoffIn TotalRunoffMG PeakRunoff RunoffCoeff',
    'Node Flooding Summary':'Name HoursFlooded MaxQ MaxDay MaxHr TotalFloodVol MaximumPondDepth',
    'Node Inflow Summary':'Name Type MaxLatInflow MaxTotalInflow MaxDay MaxHr LatInflowV TotalInflowV FlowBalErrorPerc XXX',
    'Node Surcharge Summary':'Name Type HourSurcharged MaxHeightAboveCrown MinDepthBelowRim',
    'Storage Volume Summary':'Name AvgVolume AvgPctFull EvapPctLoss ExfilPctLoss MaxVolume MaxPctFull MaxDay MaxFullHr MaxOutflow',
    'Node Depth Summary':'Name Type AvgDepth MaxNodeDepth MaxHGL MaxDay MaxHr MaxNodeDepthReported',
    'Link Flow Summary':'Name Type MaxQ MaxDay MaxHr MaxV MaxQPerc MaxDPerc',
    'Subcatchment Results':     'Date Time PrecipInchPerHour LossesInchPerHr RunoffCFS',
    'Node Results':             'Date Time InflowCFS FloodingCFS DepthFt HeadFt TSS TP TN',
    'Link Results':             'Date Time FlowCFS VelocityFPS DepthFt PercentFull',
}



from swmmio.defs.config import *
import shutil

#THIS STUFF IS INCOMPLETE/ MAYBE BROKEN
def insert_in_file(key, string, newfile, f=BASEMAP_PATH):

    #make a temp copy
    shutil.copyfile(newfile, )

    #start writing that thing
    key = '{}{}{}'.format('{{', key, '}}') #Django style
    with open(f, 'r') as bm:
        with open(newfile, 'wb') as newmap:
            for line in bm:
                if key in line:
                    newline = line.replace(key, string)
                    newmap.write(newline)
                    # newmap.write(geojson.dumps(FeatureCollection(geometries, crs=crs)))
                else:
                    newmap.write(line)

def insert_in_file_2(key, string, newfile):

    #start writing that thing
    key = '{}{}{}'.format('{{', key, '}}') #Django style
    print(key)
    with open(newfile, 'r') as newmap:
        for line in newmap:
            if key in line:
                newline = line.replace(key, string)
                newmap.write(newline)
                # newmap.write(geojson.dumps(FeatureCollection(geometries, crs=crs)))
            # else:
            #     newmap.write(line)



from swmmio.utils import functions as funcs
from swmmio import swmmio
from swmmio.version_control import utils as vc_utils
from swmmio.utils.dataframes import create_dataframeINP, create_dataframeBI
from swmmio.utils import text
import pandas as pd
from datetime import datetime
import os
import sys
from copy import deepcopy
if sys.version_info[0] < 3:
    from io import StringIO
else:
    from io import StringIO
problem_sections = ['[CURVES]', '[TIMESERIES]', '[RDII]', '[HYDROGRAPHS]']


class BuildInstructions(object):
    """
    similar to the INPDiff object, this object contains information used to
    generate an inp based on 'serialized' (though human readable, inp-esque)
    build instructions files. This object is meant to neatly encapsulate things.

    self.instructions attribute contains a dictionary with keys of the headers
    that have changes i.e. build instructions w.r.t baseline model
    """

    def __init__(self, build_instr_file=None):

        #create a change object for each section that is different from baseline
        self.instructions = {}
        self.metadata = {}
        if build_instr_file:
            #read the instructions and create a dictionary of Change objects
            allheaders = funcs.complete_inp_headers(build_instr_file)
            instructions = {}
            for section in allheaders['order']:
                change = INPDiff(build_instr_file=build_instr_file, section=section)
                instructions.update({section:change})

            self.instructions = instructions

            #read the meta data
            self.metadata = vc_utils.read_meta_data(build_instr_file)

    def __add__(self, other):
        bi = BuildInstructions()
        for section, change_obj in self.instructions.items():
            if section in other.instructions:
                new_change = change_obj + other.instructions[section]
                bi.instructions[section] = new_change
            else:
                #section doesn't exist in other, maintain current instructions
                bi.instructions[section] = change_obj

        for section, change_obj in other.instructions.items():
            if section not in self.instructions:
                bi.instructions[section] = change_obj


        #combine the metadata
        #deepcopy so child structures aren't linked to original
        bi.metadata = deepcopy(self.metadata)
        otherbaseline = other.metadata['Parent Models']['Baseline']
        otheralternatives = other.metadata['Parent Models']['Alternatives']
        bi.metadata['Parent Models']['Baseline'].update(otherbaseline)
        bi.metadata['Parent Models']['Alternatives'].update(otheralternatives)
        bi.metadata['Log'].update(other.metadata['Log'])

        return bi

    def __radd__(self, other):
        #this is so we can call sum() on a list of build_instructions
        if other == 0:
            return self
        else:
            return self.__add__(other)

    def save(self, dir, filename):
        """
        save the current BuildInstructions instance to file (human readable)
        """
        if not os.path.exists(dir):
            os.makedirs(dir)
        filepath = os.path.join(dir, filename)
        with open (filepath, 'w') as f:
            vc_utils.write_meta_data(f, self.metadata)
            for section, change_obj in self.instructions.items():
                section_df = pd.concat([change_obj.removed, change_obj.altered, change_obj.added])
                vc_utils.write_inp_section(f, allheaders=None, sectionheader=section,
                                           section_data=section_df, pad_top=False, na_fill='NaN')

    def build(self, baseline_dir, target_path):
        """
        build a complete INP file with the build instructions committed to a
        baseline model.
        """
        basemodel = swmmio.Model(baseline_dir)
        allheaders = funcs.complete_inp_headers(basemodel.inp.path)
        #new_inp = os.path.join(target_dir, 'model.inp')
        with open (target_path, 'w') as f:
            for section in allheaders['order']:

                #check if the section is not in problem_sections and there are changes
                #in self.instructions and commit changes to it from baseline accordingly
                if (section not in problem_sections
                    and allheaders['headers'][section] != 'blob'
                    and section in self.instructions):

                    #df of baseline model section
                    basedf = create_dataframeINP(basemodel.inp.path, section)

                    #grab the changes to
                    changes = self.instructions[section]

                    #remove elements that have alterations and or tagged for removal
                    remove_ids = changes.removed.index | changes.altered.index
                    new_section = basedf.drop(remove_ids)

                    #add elements
                    new_section = pd.concat([new_section, changes.altered, changes.added])
                else:
                    #section is not well understood or is problematic, just blindly copy
                    new_section = create_dataframeINP(basemodel.inp.path, section=section)

                #write the section
                vc_utils.write_inp_section(f, allheaders, section, new_section)

class INPDiff(object):
    """
    This object represents the 'changes' of a given section of a INP file
    with respect to another INP. Three main dataframes are attributes:
        added   --> includes elements that are new in model2 (compare to model1)
        removed --> elements that do not exist in model2, that were found to model1
        altered --> elements whose attributes have changes from model1 to model2


    """
    def __init__(self, model1=None, model2=None, section='[JUNCTIONS]', build_instr_file=None):

        if model1 and model2:
            df1 = create_dataframeINP(model1.inp.path, section)
            df2 = create_dataframeINP(model2.inp.path, section)

            #BUG -> this fails if a df1 or df2 is None i.e. if a section doesn't exist in one model
            added_ids = df2.index.difference(df1.index)
            removed_ids = df1.index.difference(df2.index)

            #find where elements were changed (but kept with same ID)
            common_ids = df1.index.difference(removed_ids) #original - removed = in common
            #both dfs concatenated, with matched indices for each element
            full_set = pd.concat([df1.loc[common_ids], df2.loc[common_ids]])
            #drop dupes on the set, all things that did not changed should have 1 row
            changes_with_dupes = full_set.drop_duplicates()
            #duplicate indicies are rows that have changes, isolate these
            changed_ids = changes_with_dupes.index.get_duplicates()

            added = df2.loc[added_ids].copy()
            added['Comment'] = 'Added'# from model {}'.format(model2.inp.path)
            added['Origin'] = model2.inp.path

            altered = df2.loc[changed_ids].copy()
            altered['Comment'] = 'Altered'# in model {}'.format(model2.inp.path)
            altered['Origin'] = model2.inp.path

            removed = df1.loc[removed_ids].copy()
            #comment out the removed elements
            #removed.index = ["; " + str(x) for x in removed.index]
            removed['Comment'] = 'Removed'# in model {}'.format(model2.inp.path)
            removed['Origin'] = model2.inp.path

            self.old = df1
            self.new = df2
            self.added = added
            self.removed = removed
            self.altered = altered

        if build_instr_file:
            #if generating from a build instructions file, do this (more efficient)
            df = create_dataframeBI(build_instr_file, section = section)

            self.added = df.loc[df['Comment'] == 'Added']
            self.removed = df.loc[df['Comment'] == 'Removed']
            self.altered = df.loc[df['Comment'] == 'Altered']

    def __add__(self, other):

        # this should be made more robust to catch conflicts
        change = INPDiff()
        change.added = self.added.append(other.added)
        change.removed = self.removed.append(other.removed)
        change.altered = self.altered.append(other.altered)

        return change

def generate_inp_from_diffs(basemodel, inpdiffs, target_dir):
    """
    create a new inp with respect to a baseline inp and changes instructed
    with a list of inp diff files (build instructions). This saves having to
    recalculate the differences of each model from the baseline whenever we want
    to combine versions.

    NOTE THIS ISN'T USED ANYWHERE. DELETE ????
    """

    #step 1 --> combine the diff/build instructions
    allheaders = funcs.complete_inp_headers(basemodel.inp.path)
    combi_build_instr_file = os.path.join(target_dir, 'build_instructions.txt')
    newinp = os.path.join(target_dir, 'new.inp')
    with open (combi_build_instr_file, 'w') as f:
        for header in allheaders['order']:
            s = ''
            section_header_written = False
            for inp in inpdiffs:
                sect_s = None
                if not section_header_written:
                    sect_s = text.extract_section_from_inp(inp, header,
                                                           cleanheaders=False,
                                                           return_string=True,
                                                           skipheaders=False)
                    section_header_written = True

                else:
                    sect_s = text.extract_section_from_inp(inp, header,
                                                           cleanheaders=False,
                                                           return_string=True,
                                                           skipheaders=True)

                if sect_s:
                    #remove the extra space between data in the same table
                    #coming from diffrent models.
                    if sect_s[-2:] == '\n\n':   #NOTE Check this section...
                        s += sect_s[:-1]
                    else:
                        s += sect_s

            f.write(s + '\n')

    #step 2 --> clean up the new combined diff instructions
    # df_dict = clean_inp_diff_formatting(combi_build_instr_file) #makes more human readable

    #step 3 --> create a new inp based on the baseline, with the inp_diff
    #instructions applied
    with open (newinp, 'w') as f:
        for section in allheaders['order']:
            print(section)
            if section not in problem_sections and allheaders['headers'][section] != 'blob':
                #check if a changes from baseline spreadheet exists, and use this
                #information if available to create the changes array
                df = create_dataframeINP(basemodel.inp.path, section)
                df['Origin'] = '' #add the origin column if not there
                if section in df_dict:
                    df_change = df_dict[section]
                    ids_to_drop = df_change.loc[df_change['Comment'].isin(['Removed', 'Altered'])].index
                    df = df.drop(ids_to_drop)
                    df = df.append(df_change.loc[df_change['Comment'].isin(['Added', 'Altered'])])
                new_section = df
            else:
                #blindly copy this section from the base model
                new_section = create_dataframeINP(basemodel.inp.path, section=section)

            #write the section into the inp file and the excel file
            vc_utils.write_inp_section(f, allheaders, section, new_section)



def create_inp_build_instructions(inpA, inpB, path, filename, comments=''):

    """
    pass in two inp file paths and produce a spreadsheet showing the differences
    found in each of the INP sections. These differences should then be used
    whenever we need to rebuild this model from the baseline reference model.


    Note: this should be split into a func that creates a overall model "diff"
    that can then be written as a BI file or used programmatically
    """

    allsections_a = funcs.complete_inp_headers(inpA)
    modela = swmmio.Model(inpA)
    modelb = swmmio.Model(inpB)

    #create build insructions folder
    if not os.path.exists(path):
        os.makedirs(path)
    filepath = os.path.join(path, filename) + '.txt'
    # xlpath = os.path.join(path, filename) + '.xlsx'
    # excelwriter = pd.ExcelWriter(xlpath)
    # vc_utils.create_change_info_sheet(excelwriter, modela, modelb)

    problem_sections = ['[TITLE]', '[CURVES]', '[TIMESERIES]', '[RDII]', '[HYDROGRAPHS]']
    with open (filepath, 'w') as newf:

        #write meta data
        metadata = {
            #'Baseline Model':modela.inp.path,
            #'ID':filename,
            'Parent Models':{
                            'Baseline':{inpA:vc_utils.modification_date(inpA)},
                            'Alternatives':{inpB:vc_utils.modification_date(inpB)}
                            },
            'Log':{filename:comments}
            }
        #print metadata
        vc_utils.write_meta_data(newf, metadata)
        for section in allsections_a['order']:
            if section not in problem_sections:
                #calculate the changes in the current section
                changes = INPDiff(modela, modelb, section)
                data = pd.concat([changes.removed, changes.added, changes.altered])
                #vc_utils.write_excel_inp_section(excelwriter, allsections_a, section, data)
                vc_utils.write_inp_section(newf, allsections_a, section, data, pad_top=False, na_fill='NaN') #na fill fixes SNOWPACK blanks spaces issue


    # excelwriter.save()



from swmmio import *
from swmmio.utils.dataframes import create_dataframeBI, create_dataframeRPT, create_dataframeINP
from .core import *
import swmmio.core as swmmio
'''Python SWMM Input/Output Tools'''


VERSION_INFO = (0, 3, 4)
__version__ = '.'.join(map(str, VERSION_INFO))
__author__ = 'Adam Erispaha'
__copyright__ = 'Copyright (c) 2016'
__licence__ = ''



import subprocess
import os
import pandas as pd
from swmmio.version_control import version_control as vc
from swmmio.utils.modify_model import replace_inp_section
from swmmio.utils import dataframes
from swmmio.run_models import defs
from swmmio.swmmio import Model
from swmmio.defs.config import SWMM_ENGINE_PATH


#path to the SWMM5 Engine
# SWMM_ENGINE_PATH = r'\\PWDHQR\Data\Planning & Research\Flood Risk Management\07_Software\swmm5_22.exe'

def run_simple(inp_path, swmm_eng=SWMM_ENGINE_PATH):
    """
    run a model once as is.
    """
    print('running {} with {}'.format(inp_path, swmm_eng))
    #inp_path = model.inp.path
    rpt_path = os.path.splitext(inp_path)[0] + '.rpt'

    subprocess.call([swmm_eng, inp_path, rpt_path])

def run_hot_start_sequence(inp_path, swmm_eng=SWMM_ENGINE_PATH):

    # inp_path = model.inp.path
    model = Model(inp_path)
    rpt_path = os.path.splitext(inp_path)[0] + '.rpt'
    hotstart1 = os.path.join(model.inp.dir, model.inp.name + '_hot1.hsf')
    hotstart2 = os.path.join(model.inp.dir, model.inp.name + '_hot2.hsf')

    # if not os.path.exists(hotstart1) and not os.path.exists(hotstart2):
    #create new model inp with params to save hotstart1
    print('create new model inp with params to save hotstart1')
    s = pd.Series(['SAVE HOTSTART "{}"'.format(hotstart1)])
    hot1_df = pd.DataFrame(s, columns=['[FILES]'])
    model = replace_inp_section(model.inp.path, '[FILES]', hot1_df)
    model = replace_inp_section(model.inp.path, '[REPORT]', defs.REPORT_none)
    model = replace_inp_section(model.inp.path, '[OPTIONS]', defs.OPTIONS_no_rain)
    subprocess.call([swmm_eng, model.inp.path, rpt_path])

    # if os.path.exists(hotstart1) and not os.path.exists(hotstart2):
    #create new model inp with params to use hotstart1 and save hotstart2
    print('with params to use hotstart1 and save hotstart2')
    s = pd.Series(['USE HOTSTART "{}"'.format(hotstart1), 'SAVE HOTSTART "{}"'.format(hotstart2)])
    hot2_df = pd.DataFrame(s, columns=['[FILES]'])
    model = replace_inp_section(model.inp.path, '[FILES]', hot2_df)
    subprocess.call([swmm_eng, model.inp.path, rpt_path])

    # if os.path.exists(hotstart2):
    #create new model inp with params to use hotstart2 and not save anything
    print('params to use hotstart2 and not save anything')
    s = pd.Series(['USE HOTSTART "{}"'.format(hotstart2)])
    hot3_df = pd.DataFrame(s, columns=['[FILES]'])

    model = replace_inp_section(model.inp.path, '[FILES]', hot3_df)
    model = replace_inp_section(model.inp.path, '[REPORT]', defs.REPORT_none)# defs.REPORT_nodes_links)
    model = replace_inp_section(model.inp.path, '[OPTIONS]', defs.OPTIONS_normal)

    subprocess.call([swmm_eng, model.inp.path, rpt_path])



from swmmio.utils import text as txt
from swmmio.utils import functions as funcs
import pandas as pd
import os


def create_dataframeBI(bi_path, section='[CONDUITS]'):
    """
    given a path to a biuld instructions file, create a dataframe of data in a
    given section
    """
    headerdefs = funcs.complete_inp_headers(bi_path)
    headerlist = headerdefs['headers'][section].split() + [';', 'Comment', 'Origin']
    tempfilepath = txt.extract_section_from_inp(bi_path, section,
                                                headerdefs=headerdefs,
                                                skipheaders=True)
    df = pd.read_table(tempfilepath, header=None, delim_whitespace=True,
                       skiprows=[0], index_col=0, names=headerlist, comment=None)

    os.remove(tempfilepath)  # clean up

    return df


def create_dataframeINP(inp_path, section='[CONDUITS]', ignore_comments=True,
                        comment_str=';', comment_cols=True):
    """
    given a path to an INP file, create a dataframe of data in the given
    section.
    """

    # find all the headers and their defs (section title with cleaned one-liner column headers)
    headerdefs = funcs.complete_inp_headers(inp_path)
    # create temp file with section isolated from inp file
    tempfilepath = txt.extract_section_from_inp(inp_path, section,
                                                headerdefs=headerdefs,
                                                ignore_comments=ignore_comments)
    if ignore_comments:
        comment_str = None
    if not tempfilepath:
        # if this head (section) was not found in the textfile, return a
        # blank dataframe with the appropriate schema
        print('header "{}" not found in "{}"'.format(section, inp_path))
        print('returning empty dataframe')
        headerlist = headerdefs['headers'].get(section, 'blob').split() + [';', 'Comment', 'Origin']
        blank_df = pd.DataFrame(data=None, columns=headerlist).set_index(headerlist[0])
        return blank_df

    if headerdefs['headers'][section] == 'blob':
        # return the whole row, without specifc col headers
        df = pd.read_table(tempfilepath, delim_whitespace=False, comment=comment_str)
    elif section == '[CURVES]' or section == '[TIMESERIES]':
        # return the whole row, without specifc col headers
        df = pd.read_table(tempfilepath, delim_whitespace=False)  # , index_col=0)#, skiprows=[0])
    else:
        # this section header is recognized and will be organized into known columns
        headerlist = headerdefs['headers'][section].split()
        if comment_cols:
            headerlist = headerlist + [';', 'Comment', 'Origin']
        df = pd.read_table(tempfilepath, header=None, delim_whitespace=True, skiprows=[0],
                           index_col=0, names=headerlist, comment=comment_str)

        if comment_cols:
            # add new blank comment column after a semicolon column
            df[';'] = ';'

    os.remove(tempfilepath)

    return df.rename(index=str)


def get_link_coords(row, nodexys, verticies):
    """for use in an df.apply, to get coordinates of a conduit/link """

    # cast IDs to string
    inlet_id = str(row.InletNode)
    outlet_id = str(row.OutletNode)
    xys_str = nodexys.rename(index=str)

    x1 = round(xys_str.at[inlet_id, 'X'], 4)
    y1 = round(xys_str.at[inlet_id, 'Y'], 4)
    x2 = round(xys_str.at[outlet_id, 'X'], 4)
    y2 = round(xys_str.at[outlet_id, 'Y'], 4)
    if None in [x1, x2, y1, y2]:
        print(row.name, 'problem, no coords')
    # grab any extra verts, place in between up/dwn nodes
    res = [(x1, y1)]
    if row.name in verticies.index:
        xs = verticies.loc[row.name, 'X'].tolist()
        ys = verticies.loc[row.name, 'Y'].tolist()
        if isinstance(xs, list) and isinstance(ys, list):
            # if more than one vert for this link exists, arrays are returned
            # from verticies.get_value(). it then needs to be zipped up
            res = res + list(zip(xs, ys))
        else:
            res = res + [(xs, ys)]

    res = res + [(x2, y2)]

    return [res]  # nest in a list to force a series to be returned in a df.apply


def create_dataframeRPT(rpt_path, section='Link Flow Summary', element_id=None):
    """
    given a path to an RPT file, create a dataframe of data in the given
    section.
    """

    # find all the headers and their defs (section title with cleaned one-liner column headers)
    headerdefs = funcs.complete_rpt_headers(rpt_path)
    # create temp file with section isolated from rpt file
    tempfilepath = txt.extract_section_from_rpt(rpt_path, section,
                                                headerdefs=headerdefs,
                                                element_id=element_id)

    if not tempfilepath:
        print('header "{}" not found in "{}"'.format(section, rpt_path))
        return None

    if headerdefs['headers'][section] == 'blob':
        # return the whole row, without specifc col headers
        df = pd.read_table(tempfilepath, delim_whitespace=False, comment=";")
    else:
        if element_id:
            # we'retrying to pull a time series, parse the datetimes by
            # concatenating the Date Time columns (cols 1,2)
            df0 = pd.read_table(tempfilepath, delim_whitespace=True)
            df = df0[df0.columns[2:]]  # the data sans date time columns
            df.index = pd.to_datetime(df0['Date'] + ' ' + df0['Time'])
            df.index.name = "".join(df0.columns[:2])
        else:
            # this section header is recognized, will be organized into known cols
            df = pd.read_table(tempfilepath, delim_whitespace=True, index_col=0)

    os.remove(tempfilepath)

    return df



#PARCEL FLOOD DAMAGE CALCULATIONS
#ASSUMES INPUT DATA IN ADDITION TO SWMM results


import pandas as pd

def flood_duration(node_flood_df, parcel_node_df=None,
                   parcel_node_join_csv=None, threshold=0.08333):

    """
    Given a dataframe with node flood duration and a csv table resulting
    from a one-to-many join of model shed drainage areas to parcels, this
    function returns a dataframe with flood data associated to each node.

    Assumptions:
        Flooding that occurs at any node in the SWMM model is applied to all
        parcels falling within that node's drainage area. Drainage areas are
        assumed to be generated using a Thiessen polygon method, or similar.
    """

    #read the one-to-many parcels-nodes table into a Dataframe if csv provided
    if parcel_node_df is None:
        parcel_node_df = pd.read_csv(parcel_node_join_csv)

    useful_cols = ['PARCELID', 'OUTLET', 'SUBCATCH', 'ADDRESS']
    parcel_node_df = parcel_node_df[useful_cols]

    #clean up the nodes df, using only a few columns
    useful_cols = ['HoursFlooded', 'TotalFloodVol', 'MaxHGL', 'MaxNodeDepth']
    node_flood_df = node_flood_df[useful_cols]

    #join flood data to parcels by outlet, clean a bit more of the cols
    parcel_flood = pd.merge(parcel_node_df, node_flood_df,
                            left_on='OUTLET', right_index=True)
    parcel_flood = parcel_flood[['PARCELID','HoursFlooded','TotalFloodVol']]

    #groupby parcel id to aggregate the duplicates, return the max of all dups
    parcel_flood_max = parcel_flood.groupby('PARCELID').max()

    #filter only parcels with flood duration above the threshold
    parcel_flood_max = parcel_flood_max.loc[parcel_flood_max.HoursFlooded>=threshold]
    return parcel_flood_max

def compare_flood_duration(basedf,altdf,threshold=0.08333,delta_threshold=0.25):

    df = basedf.join(altdf, lsuffix='Baseline', rsuffix='Proposed', how='outer')
    df = df.fillna(0) #any NaN means no flooding observed
    delta = df.HoursFloodedProposed - df.HoursFloodedBaseline
    df = df.assign(DeltaHours=delta)

    def categorize(parcel):
        #rename for logic clarity
        existing_flood_duration = parcel.HoursFloodedBaseline
        proposed_flood_duration = parcel.HoursFloodedProposed
        flood_duration_delta = parcel.DeltaHours

        if ((existing_flood_duration >= threshold)
            and (proposed_flood_duration >= threshold)):
            #parcel used to and still floods, check how it changed:
            if flood_duration_delta > delta_threshold:
                #flood duration increased (more than delta_threhold)
                return 'increased_flooding'

            elif flood_duration_delta < - delta_threshold:
                #flooding duration decreased (more than delta_threhold)
                return 'decreased_flooding'

        elif (existing_flood_duration < threshold
              and proposed_flood_duration >= threshold
              and abs(flood_duration_delta) >= delta_threshold):
            #flooding occurs where it perviously did not
            return 'new_flooding'

        elif (existing_flood_duration >= threshold
              and proposed_flood_duration < threshold
              and abs(flood_duration_delta) >= delta_threshold):
            #parcel that previously flooded no longer does
            return 'eliminated_flooding'

    cats = df.apply(lambda row: categorize(row), axis=1)
    df = df.assign(Category=cats)

    return df



from swmmio.defs.config import PARCEL_FEATURES, GEODATABASE
from constants import *

font_file = r"C:\Data\Code\Fonts\Raleway-Regular.ttf"
basemap_options = {
'gdb': GEODATABASE,
'features': [
    #this is an array so we can control the order of basemap layers
    {
        'feature': 'PhiladelphiaParks',
        'fill': park_green,
        'cols': ["OBJECTID"]#, "SHAPE@"]
    },
    {
        'feature': 'HydroPolyTrim',
        'fill':water_grey,
        'cols': ["OBJECTID"]#, "SHAPE@"]
    },
    {
        'feature': 'Streets_Dissolved5_SPhilly',
        'fill': lightgrey,
        'fill_anno': grey,
        'cols': ["OBJECTID", "ST_NAME"] #"SHAPE@",
    }
  ],
}



"""
Utility functions related to comparing SWMM models
"""
import pandas as pd
from swmmio.version_control.inp import INPDiff
from swmmio.utils.dataframes import create_dataframeINP
import os

def conduits_cost_estimate(conduit_df, additional_costs=None):


    def calc_area(row):
        """calculate the cross-sectional area of a sewer segment. If the segment
        is multi-barrel, the area will reflect the total of all barrels"""

        if row.Shape == 'CIRCULAR':
            d = row.Geom1
            area = 3.1415 * (d * d) / 4
            return round((area * row.Barrels),2)

        if 'RECT' in row.Shape:
            #assume triangular bottom sections (geom3) deepens the excavated box
            return (row.Geom1 + row.Geom3) * float(row.Geom2) * row.Barrels
        if row.Shape =='EGG':
            #assume geom1 is the span
            return row.Geom1*1.5 * row.Barrels

    def get_unit_cost(row):
        cost_dict = {1.0:570,
                    1.5:570,
                    1.75:610,
                    2.0:680,
                    2.25:760,
                    2.5:860,
                    3.0:1020,
                    3.5:1200,
                    4.0:1400,
                    4.5:1550,
                    5.0:1700,
                    5.5:1960,
                    6.0:2260,
                    7.0:2600,
                    7.5:3000,
                    }

        def round_to(n, precision):
            correction = 0.5 if n >= 0 else -0.5
            return int( n/precision+correction ) * precision

        def round_to_05(n):
            return round_to(n, 0.05)

        cleaned_geom = round_to_05(row.Geom1)
        RectBox_UnitCost = 80
        if row.Shape == 'CIRCULAR':
            try:
                val = cost_dict[cleaned_geom]
            except:
                val = 0
            return val
        if 'RECT' in row.Shape:
            """
            assume any triangular bottom section adds to
            overall excavation box section
            """
            val = round(RectBox_UnitCost*row.XArea,2)
            return val
        if row.Shape == 'EGG':
            """
            We're going to treat this like a box"
            """
            val = round(RectBox_UnitCost*row.XArea,2)
            return val


    def compute_conduit_cost(row):
        return row.UnitCostLF * row.Length

    def compute_volume(row):
        return row.XArea * row.Length

    def added_cost(row):
        #add in any additional cost data (from crossing water mains, etc)
        return row.CostEstimate + row.AdditionalCost

    conduit_df['XArea'] = conduit_df.apply (lambda r: calc_area (r), axis=1)
    conduit_df['UnitCostLF'] = conduit_df.apply(lambda r: get_unit_cost(r), axis=1)
    conduit_df['Volume'] = conduit_df.apply (lambda r:compute_volume (r), axis=1)
    conduit_df['CostEstimate'] = conduit_df.apply (lambda r:compute_conduit_cost (r),axis=1)

    if additional_costs:
        #read in the supplemental cost data from the csv
        addcosts = pd.read_csv(additional_costs, index_col=0)
        conduit_df = conduit_df.join(addcosts).fillna(0)
        conduit_df['TotalCostEstimate'] = conduit_df.apply (lambda r:
                                                              added_cost(r),
                                                              axis=1)
    else:
        # NOTE this lingo here is weak...
        #additional_costs not provided, rename the CostEstimate to TotalCostEstimate
        conduit_df = conduit_df.rename(columns={"CostEstimate": "TotalCostEstimate"})

    return conduit_df



# Standard library imports
import os
import json

_DEFS_PATH = os.path.abspath(os.path.dirname(__file__))
_HEADERS_JSON = os.path.join(_DEFS_PATH, 'section_headers.json')

with open(_HEADERS_JSON, 'r') as f:
    HEADERS = json.load(f)



"""
Functions to produce animations of SWMM results. Not working currently - need to
bring the Link / Node object usage up to speed
"""

from swmmio.graphics import swmm_graphics as sg
from swmmio.graphics import draw_utils as du
from swmmio.utils import swmm_utils as su
from swmmio.vendor.images2gif import writeGif
from PIL import Image, ImageDraw, ImageFont
import shutil
from datetime import datetime, timedelta
from time import strftime
import glob
import os
import pickle

def animateModel(model, startDtime=None, endDtime=None, **kwargs):

	"""
	FUNCTION NOT CURRENTLY SUPPORTED ! :(
	"""
	#unpack and update the options
	ops = du.default_draw_options()
	for key, value in kwargs.items():
		ops.update({key:value})
	#return ops
	width = ops['width']
	xplier = ops['xplier']
	bbox = ops['bbox']
	imgName = ops['imgName'] # for some reason saveImage() won't take the dict reference

	#for antialiasing, double size for now
	xplier *= width/1024 #scale the symbology sizes
	width = width*2

	#parse out the main object of this model
	inp = model.inp
	rpt = model.rpt

	#organize relavant data from SWMM files
	conduitData = model.organizeConduitData(bbox) #dictionary of overall model data, dimension, and the conduit dicts
	conduitDicts = conduitData['conduit_objects']
	pixelData = su.convertCoordinatesToPixels(conduitDicts, targetImgW=width, bbox=bbox)
	shiftRatio = pixelData['shiftRatio']
	imgSize = pixelData['imgSize']
	#node data
	nodeData = model.organizeNodeData(bbox)
	nodeDicts = nodeData['node_objects']
	su.convertCoordinatesToPixels(nodeDicts, targetImgW=width, bbox=bbox)

	#grab start and end simulation time if no range provided -> this will be huge if you're not careful!
	if not startDtime: startDtime = rpt.simulationStart
	if not endDtime: endDtime = rpt.simulationEnd

	#make sure dates are valid (within range)
	simStartDT = datetime.strptime(rpt.simulationStart, "%b-%d-%Y %H:%M:%S") #lower bound of time
	simEndDT = datetime.strptime(rpt.simulationEnd, "%b-%d-%Y %H:%M:%S") #upper bound of time
	if startDtime and endDtime:
		userStartDT = 	datetime.strptime(startDtime, "%b-%d-%Y %H:%M:%S")
		userEndDT = 	datetime.strptime(endDtime, "%b-%d-%Y %H:%M:%S")
		timeStepMod = userStartDT.minute % rpt.timeStepMin
		if userStartDT < simStartDT or userEndDT > simEndDT or timeStepMod != 0 or userEndDT < userStartDT:
			#user has entered fault date times either by not being within the
			#availble data in the rpt or by starting at something that doesn't fit the timestep
			print("PROBLEM WITH DATETIME ENTERED. Make sure it fits within data and start time rest on factor of timestep in minutes.")
			print("userStartDT = ", userStartDT, "\nuserEndDT = ", userEndDT, "\nsimStartDT = ", simStartDT, "\nsimEndDT = ", simEndDT, "\nTIMESTEP = ", rpt.timeStepMin)
			return None

	currentT = datetime.strptime(startDtime, "%b-%d-%Y %H:%M:%S") #SWMM dtime format needed
	endDtime = datetime.strptime(endDtime, "%b-%d-%Y %H:%M:%S") #SWMM dtime format needed
	delta = timedelta(minutes=rpt.timeStepMin) #SWMM reporting time step (or step to animate with)


	#use or create working dir
	if not os.path.exists( os.path.join(inp.dir, "img") ):
		os.makedirs(os.path.join(inp.dir, "img"))
	imgDir = os.path.join(inp.dir, "img")

	byteLocDictionaryFName = os.path.join(imgDir, inp.name) + "_rpt_key.txt"
	if not os.path.isfile(byteLocDictionaryFName):

		#this is a heavy operation, allow a few minutes
		print("generating byte dictionary...")
		#conduitByteLocationDict = rpt.createByteLocDict("Link Results")
		rpt.createByteLocDict("Link Results")
		rpt.createByteLocDict("Node Results")

		#save dict to disk
		dictSaveFile =  open (byteLocDictionaryFName, 'w')
		pickle.dump(rpt.elementByteLocations, dictSaveFile)
		dictSaveFile.close()

	else:
		#print "loading byte dict"
		rpt.elementByteLocations = pickle.load( open(byteLocDictionaryFName, 'r') )
		#rpt.byteLocDict = conduitByteLocationDict

	print("Started Drawing at " + strftime("%b-%d-%Y %H:%M:%S"))
	log = "Started Drawing at " + strftime("%b-%d-%Y %H:%M:%S") + "\n\nErrors:\n\n"
	drawCount = 0
	conduitErrorCount = 0

	font = ImageFont.truetype(su.fontFile, 30)
	basemapFeatureDicts = sg.createFeaturesDict(options=ops['basemap'],  bbox=bbox, shiftRatio=shiftRatio, width=width) #will be populated after first frame is produced
	while currentT <= endDtime:

		img = Image.new('RGB', imgSize, ops['bg'])
		draw = ImageDraw.Draw(img)
		currentTstr = currentT.strftime("%b-%d-%Y %H:%M:%S").upper()
		#print 'time =', currentTstr

		#DRAW THE BASEMAP
		if ops['basemap']:
			sg.draw_basemap(draw, img=img, options=ops['basemap'], width=width, bbox=bbox, featureDicts=basemapFeatureDicts, xplier = xplier)

		#DRAW THE CONDUITS
		if ops['conduitSymb']:
			for id, conduit in conduitDicts.items():
				#coordPair = coordPairDict['coordinates']
				if conduit.coordinates: #this prevents draws if no flow is supplied (RDII and such)

					su.drawConduit(conduit, coordPairDict, draw, options=ops['conduitSymb'],  rpt=rpt, dTime = currentTstr,
									xplier = xplier)

					drawCount += 1

				if drawCount > 0 and drawCount % 2000 == 0: print(str(drawCount) + " pipes drawn - simulation time = " + currentTstr)

		#DRAW THE NODES
		if ops['nodeSymb']:
			for id, node in nodeDicts.items():
				if node.coordinates: #this prevents draws if no flow is supplied (RDII and such)
					su.drawNode(node, nodeDict, draw, rpt=rpt, dTime=currentTstr, options=ops['nodeSymb'], xplier=xplier)
					drawCount += 1


		#DRAW THE ANNOTATION
		dtime = currentT.strftime("%b%d%Y_%H%M").upper()
		#su.drawAnnotation (draw, inp, rpt, imgWidth=width, title=None, currentTstr = currentTstr, fill=su.black)
		su.annotateMap (draw, model, options=ops, currentTstr = currentTstr)
		currentT += delta
		del draw


		#shrink for antialiasing niceness (though this blows up the file about 5x)
		tempImgDir = os.path.join(imgDir, "temp_frames")
		sg.saveImage(img, model, dtime, imgDir=tempImgDir, open=False, verbose=False)

	#WRITE THE GIF
	frames = []
	for image in glob.glob1(tempImgDir, "*.png"):
		imgPath = os.path.join(tempImgDir, image)
		frames.append(Image.open(imgPath))

	print("building gif with " + str(len(glob.glob1(tempImgDir, "*.png"))) + " frames...")
	if not imgName: imgName = inp.name
	gifFile = os.path.join(imgDir, imgName) + ".gif"
	frameDuration = 1.0 / float(ops['fps'])
	writeGif(gifFile, frames, duration=frameDuration)

	shutil.rmtree(tempImgDir) #delete temporary frames directory after succesful GIF

	log += "Completed drawing at " + strftime("%Y%m%d %H:%M:%S")
	with open(os.path.join(imgDir, "log.txt"), 'w') as logFile:
		logFile.write(log)

	print("Draw Count =" + str(drawCount))
	print("Video saved to:\n\t" + gifFile)

	os.startfile(gifFile)#this doesn't seem to work



from swmmio.tests.data import MODEL_FULL_FEATURES__NET_PATH
from swmmio import Model
from swmmio.elements import ModelSection
from swmmio.utils import functions
import pytest


@pytest.fixture
def test_model():
    return Model(MODEL_FULL_FEATURES__NET_PATH)


def test_model_section(test_model):
    group = ModelSection(test_model, 'junctions')
    print(group)

    bayside = Model(MODEL_FULL_FEATURES__NET_PATH)

    a = bayside.inp.junctions
    print(a)
    # tsb_ids = [1213, 13131, 232131, 12313]
    # tsbs = bayside.conduits(data=['MaxDepth, MaxQ', 'geometry'])


def test_complete_headers(test_model):
    headers = functions.complete_inp_headers(test_model.inp.path)
    sections_in_inp = [
        '[TITLE]', '[OPTIONS]', '[EVAPORATION]', '[RAINGAGES]', '[SUBCATCHMENTS]', '[SUBAREAS]', '[INFILTRATION]',
        '[JUNCTIONS]', '[OUTFALLS]', '[STORAGE]', '[CONDUITS]', '[PUMPS]', '[WEIRS]', '[XSECTIONS]', '[INFLOWS]',
        '[CURVES]', '[TIMESERIES]', '[REPORT]', '[TAGS]', '[MAP]', '[COORDINATES]', '[VERTICES]', '[Polygons]',
        '[SYMBOLS]'
    ]
    assert (all(section in headers['headers'] for section in sections_in_inp))



from swmmio.defs.config import ROOT_DIR
from swmmio.tests.data import MODEL_FULL_FEATURES_XY
import json
import pandas as pd
from geojson import Point, LineString, Polygon, FeatureCollection, Feature
import os
import shutil


def change_crs(series, in_crs, to_crs):
    """
    Change the projection of a series of coordinates
    :param series:
    :param to_crs:
    :param in_crs:
    :return: series of reprojected coordinates
    >>> import swmmio
    >>> m = swmmio.Model(MODEL_FULL_FEATURES_XY)
    >>> proj4_str = '+proj=tmerc +lat_0=36.16666666666666 +lon_0=-94.5 +k=0.9999411764705882 +x_0=850000 +y_0=0 +datum=NAD83 +units=us-ft +no_defs' #"+init=EPSG:102698"
    >>> m.crs = proj4_str
    >>> nodes = m.nodes()
    >>> change_crs(nodes['coords'], proj4_str, "+init=EPSG:4326")
    Name
    J3    [(39.236286854940964, -94.64346373821752)]
    1      [(39.23851590020802, -94.64756446847099)]
    2       [(39.2382157223383, -94.64468629488778)]
    3      [(39.23878251491925, -94.64640342340165)]
    4     [(39.238353081411915, -94.64603818939938)]
    5      [(39.23797714290924, -94.64589184224722)]
    J2     [(39.23702605103406, -94.64543916929885)]
    J4     [(39.23633648359375, -94.64190240294558)]
    J1     [(39.23723558954326, -94.64583338271147)]
    Name: coords, dtype: object
    """
    try:
        import pyproj
    except ImportError:
        raise ImportError('pyproj module needed. get this package here: ',
                          'https://pypi.python.org/pypi/pyproj')

    # SET UP THE TO AND FROM COORDINATE PROJECTION
    in_proj = pyproj.Proj(in_crs, preserve_units=True)
    to_proj = pyproj.Proj(to_crs)

    # convert coords in coordinates, vertices, and polygons inp sections
    # transform to the typical 'WGS84' coord system
    def get_xys(xy_row):
        # need to reverse to lat/long after conversion
        return [pyproj.transform(in_proj, to_proj, x, y) for x, y in xy_row]

    if isinstance(series, pd.Series):
        return series.apply(lambda row: get_xys(row))
    if isinstance(series, pd.DataFrame):
        zipped_coords = list(zip(series.X, series.Y))
        df = pd.DataFrame(data=get_xys(zipped_coords), columns=["X", "Y"], index=series.index)
        return df
    elif isinstance(series, (list, tuple)):
        if isinstance(series[0], (list, tuple)):
            return get_xys(series)
        else:
            return get_xys([series])


def coords_series_to_geometry(coords, geomtype='linestring', format='geojson'):
    """
    Convert a series of coords (list of list(s)) to a series of geometry objects.
    :param coords: series of lists of xy coordinates
    :param format: format of geometry objects to be created ('geojson', 'shapely')
    :return: series of geometry objects
    >>> import swmmio
    >>> model = swmmio.Model(MODEL_FULL_FEATURES_XY)
    >>> nodes = model.nodes()
    >>> geoms = coords_series_to_geometry(nodes['coords'], geomtype='point')
    >>> geoms.iloc[0]
    {"coordinates": [2748073.3060000003, 1117746.087], "type": "Point"}
    """

    # detect whether LineString or Point should be used
    geomtype = geomtype.lower()
    if geomtype == 'linestring':
        geoms = [LineString(latlngs) for latlngs in coords]
    elif geomtype == 'point':
        geoms = [Point(latlngs[0]) for latlngs in coords]
    elif geomtype == 'polygon':
        geoms = [Polygon([latlngs]) for latlngs in coords]

    if format.lower() == 'shape':
        # convert to shapely objects
        try:
            from shapely.geometry import shape
        except ImportError:
            raise ImportError('shapely module needed. Install it via GeoPandas with conda: ',
                              'conda install geopandas')
        geoms = [shape(g) for g in geoms]

    return pd.Series(index=coords.index, name='geometry', data=geoms)


def write_geojson(df, filename=None, geomtype='linestring', inproj='epsg:2272'):
    try:
        import pyproj
    except ImportError:
        raise ImportError('pyproj module needed. get this package here: ',
                          'https://pypi.python.org/pypi/pyproj')

    # SET UP THE TO AND FROM COORDINATE PROJECTION
    pa_plane = pyproj.Proj(init=inproj, preserve_units=True)
    wgs = pyproj.Proj(proj='longlat', datum='WGS84', ellps='WGS84')  # google maps, etc

    # CONVERT THE DF INTO JSON
    df['Name'] = df.index  # add a name column (we wont have the index)
    records = json.loads(df.to_json(orient='records'))

    # ITERATE THROUGH THE RECORDS AND CREATE GEOJSON OBJECTS
    features = []
    for rec in records:

        coordinates = rec['coords']
        del rec['coords']  # delete the coords so they aren't in the properties

        # transform to the typical 'WGS84' coord system
        latlngs = [pyproj.transform(pa_plane, wgs, *xy) for xy in coordinates]

        if geomtype == 'linestring':
            geometry = LineString(latlngs)
        elif geomtype == 'point':
            geometry = Point(latlngs)
        elif geomtype == 'polygon':
            geometry = Polygon([latlngs])

        feature = Feature(geometry=geometry, properties=rec)
        features.append(feature)

    if filename is not None:
        with open(filename, 'wb') as f:
            f.write(json.dumps(FeatureCollection(features)))
        return filename

    else:
        return FeatureCollection(features)


def write_shapefile(df, filename, geomtype='line', prj=None):
    """
    create a shapefile given a pandas Dataframe that has coordinate data in a
    column called 'coords'.
    """

    import shapefile
    df['Name'] = df.index

    # create a shp file writer object of geom type 'point'
    if geomtype == 'point':
        w = shapefile.Writer(shapefile.POINT)
    elif geomtype == 'line':
        w = shapefile.Writer(shapefile.POLYLINE)
    elif geomtype == 'polygon':
        w = shapefile.Writer(shapefile.POLYGON)

    # use the helper mode to ensure the # of records equals the # of shapes
    # (shapefile are made up of shapes and records, and need both to be valid)
    w.autoBalance = 1

    # add the fields
    for fieldname in df.columns:
        w.field(fieldname, "C")

    for k, row in df.iterrows():
        w.record(*row.tolist())
        w.line(parts=[row.coords])

    w.save(filename)

    # add projection data to the shapefile,
    if prj is None:
        # if not sepcified, the default, projection is used (PA StatePlane)
        prj = os.path.join(ROOT_DIR, 'swmmio/defs/default.prj')
    prj_filepath = os.path.splitext(filename)[0] + '.prj'
    shutil.copy(prj, prj_filepath)


def read_shapefile(shp_path):
    """
        Read a shapefile into a Pandas dataframe with a 'coords' column holding
        the geometry information. This uses the pyshp package
        """
    import shapefile

    # read file, parse out the records and shapes
    sf = shapefile.Reader(shp_path)
    fields = [x[0] for x in sf.fields][1:]
    records = sf.records()
    shps = [s.points for s in sf.shapes()]

    # write into a dataframe
    df = pd.DataFrame(columns=fields, data=records)
    df = df.assign(coords=shps)

    return df



from swmmio.version_control import version_control as vc
from swmmio.version_control import utils as vc_utils
from swmmio.utils.functions import complete_inp_headers
from swmmio import swmmio
import pandas as pd
import os

def replace_inp_section(inp_path, modified_section_header, new_data, overwrite=True):

    """
    modify an existing model by passing in new data (Pandas Dataframe)
    and the section header that should be modified. This funciton will overwrite
    all data in the old section with the passed data
    """

    tmpfilename = os.path.splitext(os.path.basename(inp_path))[0] + '_mod.inp'
    wd = os.path.dirname(inp_path)
    tmpfilepath = os.path.join(os.path.dirname(inp_path), tmpfilename)
    allheaders = complete_inp_headers(inp_path)
    basemodel = swmmio.Model(inp_path)

    with open(inp_path) as oldf:
        with open(tmpfilepath, 'w') as new:
            #create the companion excel file
            #create the MS Excel writer object
            # xlpath = os.path.join(wd, basemodel.inp.name + '_modified.xlsx')
            # excelwriter = pd.ExcelWriter(xlpath)
            # vc_utils.create_info_sheet(excelwriter, basemodel)

            #write each line as is from the original model until we find the
            #header of the section we wish to overwrite
            found_section = False
            found_next_section = False
            for line in oldf:
                if modified_section_header in line:
                    #write the replacement data in the new file now
                    vc_utils.write_inp_section(new, allheaders,
                                                modified_section_header,
                                                new_data, pad_top=False)

                    found_section = True

                if (
                    found_section and not found_next_section
                    and line.strip() in allheaders['headers']
                    and modified_section_header != line.strip()
                    ):

                    found_next_section = True
                    new.write('\n\n') #add some space before the next section

                if found_next_section or not found_section:
                    #write the lines from the original file
                    #if we haven't found the section to modify.
                    #if we have found the section and we've found the NEXT section
                    #continue writing original file's lines

                    new.write(line)

            if not found_section:
                #the header doesn't exist in the old model
                #so we should append it to the bottom of file
                vc_utils.write_inp_section(new, allheaders,
                                            modified_section_header,
                                            new_data)

    # excelwriter.save()
    #rename files and remove old if we should overwrite
    if overwrite:
        os.remove(inp_path)
        os.rename(tmpfilepath, inp_path)

    return swmmio.Model(inp_path)



from swmmio.graphics.constants import *

FLOOD_IMPACT_CATEGORIES = {
    'increased_flooding':{
        'fill':red,
    },
    'new_flooding':{
        'fill':purple,
    },
    'decreased_flooding':{
        'fill':lightblue,
    },
    'eliminated_flooding':{
        'fill':lightgreen,
    },

}
